# codes in this file are reproduced from https://github.com/microsoft/nni with some changes.
import torch
import torch.nn as nn
import torch.nn.functional as F

from . import register_nas_algo
from .base import BaseNAS
from ..space import BaseSpace
from ..utils import (
    AverageMeterGroup,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingInputChoice,
    PathSamplingLayerChoice,
)
from nni.nas.pytorch.fixed import apply_fixed_architecture
from tqdm import tqdm
from datetime import datetime
import numpy as np
from ....utils import get_logger

LOGGER = get_logger("RL_NAS")


def _get_mask(sampled, total):
    multihot = [
        i == sampled or (isinstance(sampled, list) and i in sampled)
        for i in range(total)
    ]
    return torch.tensor(multihot, dtype=torch.bool)  # pylint: disable=not-callable

class StackedLSTMCell(nn.Module):
    def __init__(self, layers, size, bias):
        super().__init__()
        self.lstm_num_layers = layers
        self.lstm_modules = nn.ModuleList(
            [nn.LSTMCell(size, size, bias=bias) for _ in range(self.lstm_num_layers)]
        )

    def forward(self, inputs, hidden):
        prev_h, prev_c = hidden
        next_h, next_c = [], []
        for i, m in enumerate(self.lstm_modules):
            curr_h, curr_c = m(inputs, (prev_h[i], prev_c[i]))
            next_c.append(curr_c)
            next_h.append(curr_h)
            # current implementation only supports batch size equals 1,
            # but the algorithm does not necessarily have this limitation
            inputs = curr_h[-1].view(1, -1)
        return next_h, next_c


class ReinforceField:
    """
    A field with ``name``, with ``total`` choices. ``choose_one`` is true if one and only one is meant to be
    selected. Otherwise, any number of choices can be chosen.
    """

    def __init__(self, name, total, choose_one):
        self.name = name
        self.total = total
        self.choose_one = choose_one

    def __repr__(self):
        return f"ReinforceField(name={self.name}, total={self.total}, choose_one={self.choose_one})"


class ReinforceController(nn.Module):
    """
    A controller that mutates the graph with RL.

    Parameters
    ----------
    fields : list of ReinforceField
        List of fields to choose.
    lstm_size : int
        Controller LSTM hidden units.
    lstm_num_layers : int
        Number of layers for stacked LSTM.
    tanh_constant : float
        Logits will be equal to ``tanh_constant * tanh(logits)``. Don't use ``tanh`` if this value is ``None``.
    skip_target : float
        Target probability that skipconnect will appear.
    temperature : float
        Temperature constant that divides the logits.
    entropy_reduction : str
        Can be one of ``sum`` and ``mean``. How the entropy of multi-input-choice is reduced.
    """

    def __init__(
        self,
        fields,
        lstm_size=64,
        lstm_num_layers=1,
        tanh_constant=1.5,
        skip_target=0.4,
        temperature=None,
        entropy_reduction="sum",
    ):
        super(ReinforceController, self).__init__()
        self.fields = fields
        self.lstm_size = lstm_size
        self.lstm_num_layers = lstm_num_layers
        self.tanh_constant = tanh_constant
        self.temperature = temperature
        self.skip_target = skip_target

        self.lstm = StackedLSTMCell(self.lstm_num_layers, self.lstm_size, False)
        self.attn_anchor = nn.Linear(self.lstm_size, self.lstm_size, bias=False)
        self.attn_query = nn.Linear(self.lstm_size, self.lstm_size, bias=False)
        self.v_attn = nn.Linear(self.lstm_size, 1, bias=False)
        self.g_emb = nn.Parameter(torch.randn(1, self.lstm_size) * 0.1)
        self.skip_targets = nn.Parameter(
            torch.tensor(
                [1.0 - self.skip_target, self.skip_target]
            ),  # pylint: disable=not-callable
            requires_grad=False,
        )
        assert entropy_reduction in [
            "sum",
            "mean",
        ], "Entropy reduction must be one of sum and mean."
        self.entropy_reduction = torch.sum if entropy_reduction == "sum" else torch.mean
        self.cross_entropy_loss = nn.CrossEntropyLoss(reduction="none")
        self.soft = nn.ModuleDict(
            {
                field.name: nn.Linear(self.lstm_size, field.total, bias=False)
                for field in fields
            }
        )
        self.embedding = nn.ModuleDict(
            {field.name: nn.Embedding(field.total, self.lstm_size) for field in fields}
        )

    def resample(self):
        self._initialize()
        result = dict()
        for field in self.fields:
            result[field.name] = self._sample_single(field)
        return result

    def _initialize(self):
        self._inputs = self.g_emb.data
        self._c = [
            torch.zeros(
                (1, self.lstm_size),
                dtype=self._inputs.dtype,
                device=self._inputs.device,
            )
            for _ in range(self.lstm_num_layers)
        ]
        self._h = [
            torch.zeros(
                (1, self.lstm_size),
                dtype=self._inputs.dtype,
                device=self._inputs.device,
            )
            for _ in range(self.lstm_num_layers)
        ]
        self.sample_log_prob = 0
        self.sample_entropy = 0
        self.sample_skip_penalty = 0

    def _lstm_next_step(self):
        self._h, self._c = self.lstm(self._inputs, (self._h, self._c))

    def _sample_single(self, field):
        self._lstm_next_step()
        logit = self.soft[field.name](self._h[-1])
        if self.temperature is not None:
            logit /= self.temperature
        if self.tanh_constant is not None:
            logit = self.tanh_constant * torch.tanh(logit)
        if field.choose_one:
            sampled = torch.multinomial(F.softmax(logit, dim=-1), 1).view(-1)
            log_prob = self.cross_entropy_loss(logit, sampled)
            self._inputs = self.embedding[field.name](sampled)
        else:
            logit = logit.view(-1, 1)
            logit = torch.cat(
                [-logit, logit], 1
            )  # pylint: disable=invalid-unary-operand-type
            sampled = torch.multinomial(F.softmax(logit, dim=-1), 1).view(-1)
            skip_prob = torch.sigmoid(logit)
            kl = torch.sum(skip_prob * torch.log(skip_prob / self.skip_targets))
            self.sample_skip_penalty += kl
            log_prob = self.cross_entropy_loss(logit, sampled)
            sampled = sampled.nonzero().view(-1)
            if sampled.sum().item():
                self._inputs = (
                    torch.sum(self.embedding[field.name](sampled.view(-1)), 0)
                    / (1.0 + torch.sum(sampled))
                ).unsqueeze(0)
            else:
                self._inputs = torch.zeros(
                    1, self.lstm_size, device=self.embedding[field.name].weight.device
                )

        sampled = sampled.detach().numpy().tolist()
        self.sample_log_prob += self.entropy_reduction(log_prob)
        entropy = (
            log_prob * torch.exp(-log_prob)
        ).detach()  # pylint: disable=invalid-unary-operand-type
        self.sample_entropy += self.entropy_reduction(entropy)
        if len(sampled) == 1:
            sampled = sampled[0]
        return sampled

@register_nas_algo("rl")
class RL(BaseNAS):
    """
    RL in GraphNas.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    device : torch.device
        ``torch.device("cpu")`` or ``torch.device("cuda")``.
    log_frequency : int
        Step count per logging.
    grad_clip : float
        Gradient clipping. Set to 0 to disable. Default: 5.
    entropy_weight : float
        Weight of sample entropy loss.
    skip_weight : float
        Weight of skip penalty loss.
    baseline_decay : float
        Decay factor of baseline. New baseline will be equal to ``baseline_decay * baseline_old + reward * (1 - baseline_decay)``.
    ctrl_lr : float
        Learning rate for RL controller.
    ctrl_steps_aggregate : int
        Number of steps that will be aggregated into one mini-batch for RL controller.
    ctrl_steps : int
        Number of mini-batches for each epoch of RL controller learning.
    ctrl_kwargs : dict
        Optional kwargs that will be passed to :class:`ReinforceController`.
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    disable_progress: boolean
        Control whether show the progress bar.
    """

    def __init__(
        self,
        num_epochs=5,
        device="auto",
        log_frequency=None,
        grad_clip=5.0,
        entropy_weight=0.0001,
        skip_weight=0.8,
        baseline_decay=0.999,
        ctrl_lr=0.00035,
        ctrl_steps_aggregate=20,
        ctrl_kwargs=None,
        n_warmup=100,
        model_lr=5e-3,
        model_wd=5e-4,
        disable_progress=False,
    ):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.log_frequency = log_frequency
        self.entropy_weight = entropy_weight
        self.skip_weight = skip_weight
        self.baseline_decay = baseline_decay
        self.baseline = 0.0
        self.ctrl_steps_aggregate = ctrl_steps_aggregate
        self.grad_clip = grad_clip
        self.ctrl_kwargs = ctrl_kwargs
        self.ctrl_lr = ctrl_lr
        self.n_warmup = n_warmup
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.disable_progress = disable_progress

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  # .to(self.device)
        self.estimator = estimator
        # replace choice
        self.nas_modules = []

        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)

        # to device
        self.model = self.model.to(self.device)
        # fields
        self.nas_fields = [
            ReinforceField(
                name,
                len(module),
                isinstance(module, PathSamplingLayerChoice) or module.n_chosen == 1,
            )
            for name, module in self.nas_modules
        ]
        self.controller = ReinforceController(
            self.nas_fields, **(self.ctrl_kwargs or {})
        )
        self.ctrl_optim = torch.optim.Adam(
            self.controller.parameters(), lr=self.ctrl_lr
        )
        # train
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                l2 = self._train_controller(i)
                bar.set_postfix(reward_controller=l2)

        selection = self.export()
        arch = space.parse_model(selection, self.device)
        # print(selection,arch)
        return arch

    def _train_controller(self, epoch):
        self.model.eval()
        self.controller.train()
        self.ctrl_optim.zero_grad()
        rewards = []
        with tqdm(
            range(self.ctrl_steps_aggregate), disable=self.disable_progress
        ) as bar:
            for ctrl_step in bar:
                self._resample()
                metric, loss = self._infer(mask="val")
                reward = metric
                bar.set_postfix(acc=metric, loss=loss.item())
                LOGGER.debug(f"{self.arch}\n{self.selection}\n{metric},{loss}")
                rewards.append(reward)
                if self.entropy_weight:
                    reward += (
                        self.entropy_weight * self.controller.sample_entropy.item()
                    )
                self.baseline = self.baseline * self.baseline_decay + reward * (
                    1 - self.baseline_decay
                )
                loss = self.controller.sample_log_prob * (reward - self.baseline)
                if self.skip_weight:
                    loss += self.skip_weight * self.controller.sample_skip_penalty
                loss /= self.ctrl_steps_aggregate
                loss.backward()

                if (ctrl_step + 1) % self.ctrl_steps_aggregate == 0:
                    if self.grad_clip > 0:
                        nn.utils.clip_grad_norm_(
                            self.controller.parameters(), self.grad_clip
                        )
                    self.ctrl_optim.step()
                    self.ctrl_optim.zero_grad()

                if (
                    self.log_frequency is not None
                    and ctrl_step % self.log_frequency == 0
                ):
                    LOGGER.debug(
                        "RL Epoch [%d/%d] Step [%d/%d]  %s",
                        epoch + 1,
                        self.num_epochs,
                        ctrl_step + 1,
                        self.ctrl_steps_aggregate,
                    )
        return sum(rewards) / len(rewards)

    def _resample(self):
        result = self.controller.resample()
        self.arch = self.model.parse_model(result, device=self.device)
        self.selection = result

    def export(self):
        self.controller.eval()
        with torch.no_grad():
            return self.controller.resample()

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.arch._model, self.dataset, mask=mask)
        return metric[0], loss



@register_nas_algo("graphnas")
class GraphNasRL(BaseNAS):
    """
    RL in GraphNas.

    Parameters
    ----------
    device : torch.device
        ``torch.device("cpu")`` or ``torch.device("cuda")``.
    num_epochs : int
        Number of epochs planned for training.
    log_frequency : int
        Step count per logging.
    grad_clip : float
        Gradient clipping. Set to 0 to disable. Default: 5.
    entropy_weight : float
        Weight of sample entropy loss.
    skip_weight : float
        Weight of skip penalty loss.
    baseline_decay : float
        Decay factor of baseline. New baseline will be equal to ``baseline_decay * baseline_old + reward * (1 - baseline_decay)``.
    ctrl_lr : float
        Learning rate for RL controller.
    ctrl_steps_aggregate : int
        Number of steps that will be aggregated into one mini-batch for RL controller.
    ctrl_steps : int
        Number of mini-batches for each epoch of RL controller learning.
    ctrl_kwargs : dict
        Optional kwargs that will be passed to :class:`ReinforceController`.
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    topk : int
        Number of architectures kept in training process.
    disable_progeress: boolean
        Control whether show the progress bar.
    """

    def __init__(
        self,
        device="auto",
        num_epochs=10,
        log_frequency=None,
        grad_clip=5.0,
        entropy_weight=0.0001,
        skip_weight=0,
        baseline_decay=0.95,
        ctrl_lr=0.00035,
        ctrl_steps_aggregate=100,
        ctrl_kwargs=None,
        n_warmup=100,
        model_lr=5e-3,
        model_wd=5e-4,
        topk=5,
        disable_progress=False,
        hardware_metric_limit=None,
    ):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.log_frequency = log_frequency
        self.entropy_weight = entropy_weight
        self.skip_weight = skip_weight
        self.baseline_decay = baseline_decay
        self.ctrl_steps_aggregate = ctrl_steps_aggregate
        self.grad_clip = grad_clip
        self.ctrl_kwargs = ctrl_kwargs
        self.ctrl_lr = ctrl_lr
        self.n_warmup = n_warmup
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.hist = []
        self.topk = topk
        self.disable_progress = disable_progress
        self.hardware_metric_limit = hardware_metric_limit
        self.allhist=[]

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  # .to(self.device)
        self.estimator = estimator
        # replace choice
        self.nas_modules = []

        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)

        # to device
        self.model = self.model.to(self.device)
        # fields
        self.nas_fields = [
            ReinforceField(
                name,
                len(module),
                isinstance(module, PathSamplingLayerChoice) or module.n_chosen == 1,
            )
            for name, module in self.nas_modules
        ]
        self.controller = ReinforceController(
            self.nas_fields,
            lstm_size=100,
            temperature=5.0,
            tanh_constant=2.5,
            **(self.ctrl_kwargs or {}),
        )
        self.ctrl_optim = torch.optim.Adam(
            self.controller.parameters(), lr=self.ctrl_lr
        )
        # train
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                l2 = self._train_controller(i)
                bar.set_postfix(reward_controller=l2)

        # selection=self.export()

        selections = [x[1] for x in self.hist]
        candidiate_accs = [-x[0] for x in self.hist]
        # print('candidiate accuracies',candidiate_accs)
        selection = self._choose_best(selections)
        arch = space.parse_model(selection, self.device)
        # print(selection,arch)
        return arch

    def _choose_best(self, selections):
        # graphnas use top 5 models, can evaluate 20 times epoch and choose the best.
        results = []
        for selection in selections:
            accs = []
            for i in tqdm(range(20), disable=self.disable_progress):
                self.arch = self.model.parse_model(selection, device=self.device)
                metric, loss, _ = self._infer(mask="val")
                accs.append(metric)
            result = np.mean(accs)
            LOGGER.info(
                "selection {} \n acc {:.4f} +- {:.4f}".format(
                    selection, np.mean(accs), np.std(accs) / np.sqrt(20)
                )
            )
            results.append(result)
        best_selection = selections[np.argmax(results)]
        return best_selection

    def _train_controller(self, epoch):
        self.model.eval()
        self.controller.train()
        self.ctrl_optim.zero_grad()
        rewards = []
        baseline = None
        # diff: graph nas train 100 and derive 100 for every epoch(10 epochs), we just train 100(20 epochs). totol num of samples are same (2000)
        with tqdm(
            range(self.ctrl_steps_aggregate), disable=self.disable_progress
        ) as bar:
            for ctrl_step in bar:
                self._resample()
                metric, loss, hardware_metric = self._infer(mask="val")
                reward = metric

                # bar.set_postfix(acc=metric,loss=loss.item())
                LOGGER.debug(f"{self.arch}\n{self.selection}\n{metric},{loss}")
                # diff: not do reward shaping as in graphnas code
                if (
                    self.hardware_metric_limit is None
                    or hardware_metric[0] < self.hardware_metric_limit
                ):
                    self.hist.append([-metric, self.selection])
                    self.allhist.append([-metric, self.selection])
                    if len(self.hist) > self.topk:
                        self.hist.sort(key=lambda x: x[0])
                        self.hist.pop()
                rewards.append(reward)

                if self.entropy_weight:
                    reward += (
                        self.entropy_weight * self.controller.sample_entropy.item()
                    )

                if not baseline:
                    baseline = reward
                else:
                    baseline = baseline * self.baseline_decay + reward * (
                        1 - self.baseline_decay
                    )

                loss = self.controller.sample_log_prob * (reward - baseline)
                self.ctrl_optim.zero_grad()
                loss.backward()

                self.ctrl_optim.step()

                bar.set_postfix(acc=metric, max_acc=max(rewards))
        
        LOGGER.info("epoch:{}, mean rewards:{}".format(epoch, sum(rewards) / len(rewards)))
        return sum(rewards) / len(rewards)

    def _resample(self):
        result = self.controller.resample()
        self.arch = self.model.parse_model(result, device=self.device)
        self.selection = result

    def export(self):
        self.controller.eval()
        with torch.no_grad():
            return self.controller.resample()

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.arch._model, self.dataset, mask=mask)
        return metric[0], loss, metric[1:]



import torch.nn.functional
import typing as _typing
import dgl
from dgl.nn.pytorch.conv import GraphConv
from .. import base_encoder, encoder_registry
from ... import _utils


class _GCN(torch.nn.Module):
    def __init__(
            self, input_dimension: int,
            dimensions: _typing.Sequence[int],
            act: _typing.Optional[str],
            dropout: _typing.Optional[float]
    ):
        super(_GCN, self).__init__()
        self.__convolution_layers: torch.nn.ModuleList = torch.nn.ModuleList()
        for layer, _dimension in enumerate(dimensions):
            self.__convolution_layers.append(
                GraphConv(
                    input_dimension if layer == 0 else dimensions[layer - 1],
                    _dimension
                )
            )
        self._act: _typing.Optional[str] = act
        self._dropout: _typing.Optional[float] = dropout

    def forward(self, graph: dgl.DGLGraph, *__args, **__kwargs):
        x: torch.Tensor = graph.ndata['feat']
        results: _typing.MutableSequence[torch.Tensor] = [x]
        for _layer in range(len(self.__convolution_layers)):
            x = self.__convolution_layers[_layer](graph, x)
            if _layer < len(self.__convolution_layers) - 1:
                x = _utils.activation.activation_func(x, self._act)
                x = torch.nn.functional.dropout(x, self._dropout, self.training)
            results.append(x)
        return results


@encoder_registry.EncoderUniversalRegistry.register_encoder('gcn')
@encoder_registry.EncoderUniversalRegistry.register_encoder('gcn_encoder')
class GCNEncoderMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(GCNEncoderMaintainer, self).__init__(
            input_dimension, final_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "add_self_loops",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "normalize",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]
        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [128, 64],
            "dropout": 0.,
            "act": "relu",
        }

    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters["hidden"])
        if (
                self.final_dimension not in (Ellipsis, None) and
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
        self._encoder = _GCN(
            self.input_dimension, dimensions,
            self.hyper_parameters["act"],
            self.hyper_parameters["dropout"]
        ).to(self.device)
        return True


import numpy as np
import random
import torch
import torch.utils.data
import typing as _typing
from sklearn.model_selection import StratifiedKFold, KFold
from autogl import backend as _backend
from autogl.data import InMemoryDataset


def index_to_mask(index: torch.Tensor, size):
    mask = torch.zeros(size, dtype=torch.bool, device=index.device)
    mask[index] = True
    return mask


def random_splits_mask(
        dataset: InMemoryDataset,
        train_ratio: float = 0.2, val_ratio: float = 0.4,
        seed: _typing.Optional[int] = None
) -> InMemoryDataset:
    r"""If the data has masks for train/val/test, return the splits with specific ratio.

    Parameters
    ----------
    dataset : InMemoryDataset
        graph set
    train_ratio : float
        the portion of data that used for training.

    val_ratio : float
        the portion of data that used for validation.

    seed : int
        random seed for splitting dataset.
    """
    if not train_ratio + val_ratio <= 1:
        raise ValueError("the sum of provided train_ratio and val_ratio is larger than 1")

    def __random_split_masks(
            num_nodes: int
    ) -> _typing.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        _rng_state: torch.Tensor = torch.get_rng_state()
        if seed is not None and isinstance(seed, int):
            torch.manual_seed(seed)
        perm = torch.randperm(num_nodes)
        train_index = perm[:int(num_nodes * train_ratio)]
        val_index = perm[int(num_nodes * train_ratio): int(num_nodes * (train_ratio + val_ratio))]
        test_index = perm[int(num_nodes * (train_ratio + val_ratio)):]
        torch.set_rng_state(_rng_state)
        return (
            index_to_mask(train_index, num_nodes),
            index_to_mask(val_index, num_nodes),
            index_to_mask(test_index, num_nodes)
        )

    for index in range(len(dataset)):
        for node_type in dataset[index].nodes:
            data_keys = [data_key for data_key in dataset[index].nodes.data]
            if len(data_keys) > 0:
                _num_nodes: int = dataset[index].nodes[node_type].data[data_keys[0]].size(0)
                _masks: _typing.Tuple[torch.Tensor, torch.Tensor, torch.Tensor] = (
                    __random_split_masks(_num_nodes)
                )
                dataset[index].nodes[node_type].data["train_mask"] = _masks[0]
                dataset[index].nodes[node_type].data["val_mask"] = _masks[1]
                dataset[index].nodes[node_type].data["test_mask"] = _masks[2]
    return dataset


def random_splits_mask_class(
        dataset: InMemoryDataset,
        num_train_per_class: int = 20,
        num_val_per_class: int = 30,
        total_num_val: _typing.Optional[int] = ...,
        total_num_test: _typing.Optional[int] = ...,
        seed: _typing.Optional[int] = ...
):
    r"""If the data has masks for train/val/test, return the splits with specific number of samples from every class for training as suggested in Pitfalls of graph neural network evaluation [#]_ for semi-supervised learning.

    References
    ----------
    .. [#] Shchur, O., Mumme, M., Bojchevski, A., & Günnemann, S. (2018).
        Pitfalls of graph neural network evaluation.
        arXiv preprint arXiv:1811.05868.

    Parameters
    ----------
    dataset: InMemoryDataset
        instance of ``InMemoryDataset``
    num_train_per_class : int
        the number of samples from every class used for training.

    num_val_per_class : int
        the number of samples from every class used for validation.

    total_num_val : int
        the total number of nodes that used for validation as alternative.

    total_num_test : int
        the total number of nodes that used for testing as alternative. The rest of the data will be seleted as test set if num_test set to None.

    seed : int
        random seed for splitting dataset.
    """
    for graph_index in range(len(dataset)):
        for node_type in dataset[graph_index].nodes:
            if (
                    'y' in dataset[graph_index].nodes[node_type].data and
                    'label' in dataset[graph_index].nodes[node_type].data
            ):
                raise ValueError(
                    f"Both 'y' and 'label' data exist "
                    f"for node type [{node_type}] in "
                    f"graph with index [{graph_index}]."
                )
            elif (
                    'y' not in dataset[graph_index].nodes[node_type].data and
                    'label' not in dataset[graph_index].nodes[node_type].data
            ):
                continue
            elif 'y' in dataset[graph_index].nodes[node_type].data:
                label: torch.Tensor = dataset[graph_index].nodes[node_type].data['y']
            elif 'label' in dataset[graph_index].nodes[node_type].data:
                label: torch.Tensor = dataset[graph_index].nodes[node_type].data['label']
            else:
                raise RuntimeError
            num_nodes: int = label.size(0)
            num_classes: int = label.cpu().max().item() + 1

            _rng_state: torch.Tensor = torch.get_rng_state()
            if seed not in (Ellipsis, None) and isinstance(seed, int):
                torch.manual_seed(seed)
            train_mask = torch.zeros(num_nodes, dtype=torch.bool, device=label.device)
            val_mask = torch.zeros(num_nodes, dtype=torch.bool, device=label.device)
            test_mask = torch.zeros(num_nodes, dtype=torch.bool, device=label.device)
            for class_index in range(num_classes):
                idx = (label == class_index).nonzero().view(-1)
                assert num_train_per_class + num_val_per_class < idx.size(0), (
                    f"the total number of samples from every class "
                    f"used for training and validation is larger than "
                    f"the total samples in class [{class_index}] for node type [{node_type}] "
                    f"in graph with index [{graph_index}]"
                )
                randomized_index: torch.Tensor = torch.randperm(idx.size(0))
                train_idx = idx[randomized_index[:num_train_per_class]]
                val_idx = idx[
                    randomized_index[num_train_per_class: (num_train_per_class + num_val_per_class)]
                ]
                train_mask[train_idx] = True
                val_mask[val_idx] = True

            if isinstance(total_num_val, int) and total_num_val > 0:
                remaining = (~train_mask).nonzero().view(-1)
                remaining = remaining[torch.randperm(remaining.size(0))]
                val_mask[remaining[:total_num_val]] = True
                if isinstance(total_num_test, int) and total_num_test > 0:
                    test_mask[remaining[total_num_val: (total_num_val + total_num_test)]] = True
                else:
                    test_mask[remaining[total_num_val:]] = True
            else:
                remaining = (~(train_mask + val_mask)).nonzero().view(-1)
                test_mask[remaining] = True

            torch.set_rng_state(_rng_state)
            dataset[graph_index].nodes[node_type].data["train_mask"] = train_mask
            dataset[graph_index].nodes[node_type].data["val_mask"] = val_mask
            dataset[graph_index].nodes[node_type].data["test_mask"] = test_mask
    return dataset


def graph_cross_validation(
        dataset: InMemoryDataset,
        n_splits: int = 10, shuffle: bool = True,
        random_seed: _typing.Optional[int] = ...,
        stratify: bool = False
) -> InMemoryDataset:
    r"""Cross validation for graph classification data

    Parameters
    ----------
    dataset : InMemoryDataset
        dataset with multiple graphs.

    n_splits : int
        the number of folds to split.

    shuffle : bool
        shuffle or not for sklearn.model_selection.StratifiedKFold

    random_seed : int
        random_state for sklearn.model_selection.StratifiedKFold

    stratify: bool
    """
    if not isinstance(n_splits, int):
        raise TypeError
    elif not n_splits > 0:
        raise ValueError
    if not isinstance(shuffle, bool):
        raise TypeError
    if not (random_seed in (Ellipsis, None) or isinstance(random_seed, int)):
        raise TypeError
    elif isinstance(random_seed, int) and random_seed >= 0:
        _random_seed: int = random_seed
    else:
        _random_seed: int = random.randrange(0, 65536)
    if not isinstance(stratify, bool):
        raise TypeError

    if stratify:
        kf = StratifiedKFold(
            n_splits=n_splits, shuffle=shuffle, random_state=_random_seed
        )
    else:
        kf = KFold(
            n_splits=n_splits, shuffle=shuffle, random_state=_random_seed
        )
    dataset_y = [g.data['y' if 'y' in g.data else 'label'].item() for g in dataset]
    idx_list = [
        (train_index.tolist(), test_index.tolist())
        for train_index, test_index
        in kf.split(np.zeros(len(dataset)), np.array(dataset_y))
    ]

    dataset.folds = idx_list
    dataset.train_index = idx_list[0][0]
    dataset.val_index = idx_list[0][1]
    return dataset


def set_fold(dataset: InMemoryDataset, fold_id: int) -> InMemoryDataset:
    r"""Set fold for graph dataset consist of multiple graphs.

    Parameters
    ----------
    dataset: `autogl.data.InMemoryDataset`
        dataset with multiple graphs.
    fold_id: `int`
        The fold in to use, MUST be in [0, dataset.n_splits)

    Returns
    -------
    `autogl.data.InMemoryDataset`
        The reference of original dataset.
    """
    if not (hasattr(dataset, 'folds') and dataset.folds is not None):
        raise ValueError("Dataset do NOT contain folds")
    if not 0 <= fold_id < len(dataset.folds):
        raise ValueError(
            f"Fold id {fold_id} exceed total cross validation split number {len(dataset.folds)}"
        )
    dataset.train_index = dataset.folds[fold_id].train_index
    dataset.val_index = dataset.folds[fold_id].val_index
    return dataset


def graph_random_splits(
        dataset: InMemoryDataset,
        train_ratio: float = 0.2,
        val_ratio: float = 0.4,
        seed: _typing.Optional[int] = ...
):
    r"""Splitting graph dataset with specific ratio for train/val/test.

    Parameters
    ----------
    dataset: ``InMemoryStaticGraphSet``

    train_ratio : float
        the portion of data that used for training.

    val_ratio : float
        the portion of data that used for validation.

    seed : int
        random seed for splitting dataset.
    """
    _rng_state = torch.get_rng_state()
    if isinstance(seed, int):
        torch.manual_seed(seed)
    perm = torch.randperm(len(dataset))
    train_index = perm[: int(len(dataset) * train_ratio)]
    val_index = (
        perm[int(len(dataset) * train_ratio): int(len(dataset) * (train_ratio + val_ratio))]
    )
    test_index = perm[int(len(dataset) * (train_ratio + val_ratio)):]
    dataset.train_index = train_index.tolist()
    dataset.val_index = val_index.tolist()
    dataset.test_index = test_index.tolist()
    torch.set_rng_state(_rng_state)
    return dataset


def graph_get_split(
        dataset, mask: str = "train",
        is_loader: bool = True, batch_size: int = 128,
        num_workers: int = 0, shuffle: bool = False
) -> _typing.Union[torch.utils.data.DataLoader, _typing.Iterable]:
    r"""Get train/test dataset/dataloader after cross validation.

    Parameters
    ----------
    dataset:
        dataset with multiple graphs.

    mask : str

    is_loader : bool
        return original dataset or data loader

    batch_size : int
        batch_size for generating Dataloader
    num_workers : int
        number of workers parameter for data loader
    shuffle: bool
        whether shuffle the dataloader
    """
    if not isinstance(mask, str):
        raise TypeError
    elif mask.lower() not in ("train", "val", "test"):
        raise ValueError
    if not isinstance(is_loader, bool):
        raise TypeError
    if not isinstance(batch_size, int):
        raise TypeError
    elif not batch_size > 0:
        raise ValueError
    if not isinstance(num_workers, int):
        raise TypeError
    elif not num_workers >= 0:
        raise ValueError

    if mask.lower() not in ("train", "val", "test"):
        raise ValueError
    elif mask.lower() == "train":
        optional_dataset_split = dataset.train_split
        if optional_dataset_split is None:
            raise ValueError(f"Provided dataset do NOT have {mask} split")
        else:
            sub_dataset = InMemoryDataset(
                optional_dataset_split, train_index=list(range(len(optional_dataset_split)))
            )
    elif mask.lower() == "val":
        optional_dataset_split = dataset.val_split
        if optional_dataset_split is None:
            raise ValueError(f"Provided dataset do NOT have {mask} split")
        else:
            sub_dataset = InMemoryDataset(
                optional_dataset_split, val_index=list(range(len(optional_dataset_split)))
            )
    elif mask.lower() == "test":
        optional_dataset_split = dataset.test_split
        if optional_dataset_split is None:
            raise ValueError(f"Provided dataset do NOT have {mask} split")
        else:
            sub_dataset = InMemoryDataset(
                optional_dataset_split, test_index=list(range(len(optional_dataset_split)))
            )
    else:
        raise ValueError(
            f"The provided mask parameter must be a str in ['train', 'val', 'test'], "
            f"illegal provided value is [{mask}]"
        )
    if not is_loader:
        return sub_dataset
    if is_loader:
        if not (_backend.DependentBackend.is_dgl() or _backend.DependentBackend.is_pyg()):
            raise RuntimeError("Unsupported backend")
        elif _backend.DependentBackend.is_dgl():
            from dgl.dataloading.pytorch import GraphDataLoader
            return GraphDataLoader(
                sub_dataset,
                **{"batch_size": batch_size, "num_workers": num_workers},
                shuffle=shuffle
            )
        elif _backend.DependentBackend.is_pyg():
            _sub_dataset: _typing.Any = optional_dataset_split
            import torch_geometric
            if int(torch_geometric.__version__.split('.')[0]) >= 2:
                # version 2.x
                from torch_geometric.loader import DataLoader
            else:
                from torch_geometric.data import DataLoader
            return DataLoader(
                _sub_dataset, batch_size=batch_size, num_workers=num_workers, shuffle=shuffle
            )
    else:
        return sub_dataset



# pyg augmentation method from <https://github.com/Shen-Lab/GraphCL/>

import random
import torch
import numpy as np
from itertools import repeat, product
from torch_geometric.data import Batch

class BaseAugmentation:
    def __init__(self, aug_ratio=None):
        self.aug_ratio = aug_ratio
    
    def _aug_data(self, data):
        pass
    
    def __call__(self, batch):
        new_data = []
        for data in batch.to_data_list():
            new_data.append(self._aug_data(data))
        return Batch.from_data_list(new_data)

class DropNode(BaseAugmentation):
    def __init__(self, aug_ratio):
        super().__init__(aug_ratio)
    
    def _aug_data(self, data):
        node_num, _ = data.x.size()
        _, edge_num = data.edge_index.size()
        drop_num = int(node_num * self.aug_ratio)

        idx_perm = np.random.permutation(node_num)

        idx_drop = idx_perm[:drop_num]
        idx_nondrop = idx_perm[drop_num:]
        idx_nondrop.sort()
        idx_dict = {idx_nondrop[n]:n for n in list(range(idx_nondrop.shape[0]))}

        edge_index = data.edge_index.numpy()
        adj = torch.zeros((node_num, node_num))
        adj[edge_index[0], edge_index[1]] = 1
        adj = adj[idx_nondrop, :][:, idx_nondrop]
        edge_index = adj.nonzero().t()

        try:
            data.edge_index = edge_index
            data.x = data.x[idx_nondrop]
        except:
            data = data
        return data

class PermuteEdge(BaseAugmentation):
    def __init__(self, aug_ratio):
        super().__init__(aug_ratio)

    def _aug_data(self, data):
        node_num, _ = data.x.size()
        _, edge_num = data.edge_index.size()
        permute_num = int(edge_num * self.aug_ratio)

        edge_index = data.edge_index.numpy()

        idx_add = np.random.choice(node_num, (2, permute_num))

        # idx_add = [[idx_add[0, n], idx_add[1, n]] for n in range(permute_num) if not (idx_add[0, n], idx_add[1, n]) in edge_index]
        # edge_index = [edge_index[n] for n in range(edge_num) if not n in np.random.choice(edge_num, permute_num, replace=False)] + idx_add

        edge_index = np.concatenate((edge_index[:, np.random.choice(edge_num, (edge_num - permute_num), replace=False)], idx_add), axis=1)
        data.edge_index = torch.tensor(edge_index)

        return data

class SubGraph(BaseAugmentation):
    def __init__(self, aug_ratio):
        super().__init__(aug_ratio)

    def _aug_data(self, data):
        node_num, _ = data.x.size()
        _, edge_num = data.edge_index.size()
        sub_num = int(node_num * self.aug_ratio)

        edge_index = data.edge_index.numpy()

        idx_sub = [np.random.randint(node_num, size=1)[0]]
        idx_neigh = set([n for n in edge_index[1][edge_index[0]==idx_sub[0]]])

        count = 0
        while len(idx_sub) <= sub_num:
            count = count + 1
            if count > node_num:
                break
            if len(idx_neigh) == 0:
                break
            sample_node = np.random.choice(list(idx_neigh))
            if sample_node in idx_sub:
                continue
            idx_sub.append(sample_node)
            idx_neigh.union(set([n for n in edge_index[1][edge_index[0]==idx_sub[-1]]]))

        idx_drop = [n for n in range(node_num) if not n in idx_sub]
        idx_nondrop = idx_sub
        data.x = data.x[idx_nondrop]
        idx_dict = {idx_nondrop[n]:n for n in list(range(len(idx_nondrop)))}

        edge_index = data.edge_index.numpy()
        adj = torch.zeros((node_num, node_num))
        adj[edge_index[0], edge_index[1]] = 1
        adj[list(range(node_num)), list(range(node_num))] = 1
        adj = adj[idx_nondrop, :][:, idx_nondrop]
        edge_index = adj.nonzero().t()

        # edge_index = [[idx_dict[edge_index[0, n]], idx_dict[edge_index[1, n]]] for n in range(edge_num) if (not edge_index[0, n] in idx_drop) and (not edge_index[1, n] in idx_drop)] + [[n, n] for n in idx_nondrop]
        data.edge_index = edge_index

        return data

class MaskNode(BaseAugmentation):
    def __init__(self, aug_ratio):
        super().__init__(aug_ratio)

    def _aug_data(self, data):
        node_num, feat_dim = data.x.size()
        mask_num = int(node_num * self.aug_ratio)

        token = data.x.mean(dim=0)
        idx_mask = np.random.choice(node_num, mask_num, replace=False)
        data.x[idx_mask] = torch.tensor(token, dtype=torch.float32)

        return data

class RandomView(BaseAugmentation):
    def __init__(self, candidates):
        super().__init__()
        self.candidates = candidates
    
    def _aug_data(self, data):
        view = random.choice(self.candidates)
        return view._aug_data(data)



from operator import xor
from .utils import data_is_tensor, data_tensor2np, data_np2tensor
import numpy as np
import copy
from torch_geometric.utils.convert import to_networkx
from torch.utils.data import Dataset
import torch
from ...utils import get_logger

LOGGER = get_logger("Feature")


class BaseFeature:
    r"""Any feature funcion object should inherit BaseFeature,
    which provides basic transformations and composing operation for feature
    engineering. Basic transformations include data type adjusting(tensor or numpy),
    complementing necessary attributes for future transform. Any subclass needs
    to overload methods ``_func`` and ``_transform`` to implement feature transformation.
    For specific needs, you may want to overload methods ``_preprocess`` and ``_postprocess``
    to enable specific processing before and after ``_transform`` .

    Parameters
    ----------
    pipe : list
        stores pipeline of ``BaseFeature``.
    data_t: str
        represents the data type needed for this transform, where 'tensor' accounts for ``torch.Tensor``,
        'np' for ``numpy.array`` and 'nx' for ``networkx``. When ``data_t`` values 'nx', then a ``networkx.DiGraph`` will
        be added to data as data.G .
    multigraph : bool
        determine whether it supports dataset with multiple graphs
    subgraph : bool
        determine whether it extracts subgraph features.
    """

    def __init__(self, pipe=None, data_t="tensor", multigraph=True, subgraph=False):
        r""""""
        if pipe is None:
            pipe = [self]
        self._pipe = pipe
        self._data_t = data_t
        self._multigraph = multigraph
        self._subgraph = subgraph

    def __and__(self, o):
        r"""enable and operation to support feature engineering pipeline syntax like
        SeFilterConstant()&GeEigen()&...
        """
        return BaseFeature(self._pipe + o._pipe)

    def _rebuild(self, dataset, datalist):
        dataset.__indices__ = None
        data, slices = dataset.collate(datalist)
        dataset.data.__dict__.update(data.__dict__)
        dataset.slices.update(slices)
        return dataset

    def _adjust_t(self, data):
        r"""adjust data type for current transform."""
        if self._data_t == "tensor":
            data_np2tensor(data)
        elif self._data_t == "np":
            data_tensor2np(data)
        elif self._data_t == "nx":
            if not hasattr(data, "G") or data.G is None:
                data.G = to_networkx(data, to_undirected=True)

    def _adjust_to_tensor(self, data):
        if self._data_t == "tensor":
            pass
        else:
            data_np2tensor(data)

    def _preprocess(self, data):
        pass

    def _postprocess(self, data):
        pass

    def _check_dataset(self, dataset):
        if len(dataset) > 1:
            for p in self._pipe:
                if not p._multigraph:
                    LOGGER.warn(p.__class__.__name__, " does not support multigraph")
                    return False
        return True

    def _fit(self, data):
        pass

    def _transform(self, data):
        return data

    def _fit_transform(self, data):
        self._fit(data)
        return self._transform(data)

    def fit(self, dataset):
        r"""fit dataset"""
        if not self._check_dataset(dataset):
            return
        dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for p in self._pipe:
                _dataset = [x for x in dataset]
                for i, datai in enumerate(_dataset):
                    p._adjust_t(datai)
                    p._preprocess(datai)
                    p._fit_transform(datai)
                    p._postprocess(datai)
                    p._adjust_to_tensor(datai)
                    _dataset[i] = datai
                dataset = self._rebuild(dataset, _dataset)

    def transform(self, dataset, inplace=True):
        r"""transform dataset inplace or not w.r.t bool argument ``inplace``"""
        if not self._check_dataset(dataset):
            return dataset
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for p in self._pipe:
                self._dataset = _dataset = [x for x in dataset]
                for i, datai in enumerate(_dataset):
                    p._adjust_t(datai)
                    p._preprocess(datai)
                    datai = p._transform(datai)
                    p._postprocess(datai)
                    p._adjust_to_tensor(datai)
                    _dataset[i] = datai
                dataset = self._rebuild(dataset, _dataset)
        dataset.data = data_np2tensor(dataset.data)
        return dataset

    def fit_transform(self, dataset, inplace=True):
        r"""fit and transform dataset inplace or not w.r.t bool argument ``inplace``"""
        self.fit(dataset)
        return self.transform(dataset, inplace=inplace)

    @staticmethod
    def compose(trans_list):
        r"""put a list of ``BaseFeature`` into feature engineering pipeline"""
        res = BaseFeature()
        for tran in trans_list:
            res = res & tran
        return res


class BaseFeatureEngineer(BaseFeature):
    def __init__(self, data_t="np", multigraph=False, *args, **kwargs):
        super(BaseFeatureEngineer, self).__init__(
            data_t=data_t, multigraph=multigraph, *args, **kwargs
        )
        self.args = args
        self.kwargs = kwargs


class TransformWrapper(BaseFeature):
    def __init__(self, cls, *args, **kwargs):
        super(TransformWrapper, self).__init__(data_t="tensor", *args, **kwargs)
        self._cls = cls
        self._func = None
        self._args = args
        self._kwargs = kwargs

    def __call__(self):
        if self._func is None:
            self._func = self._cls(*self._args, **self._kwargs)
            return self

    def _transform(self, data=None):
        return self._func(data)



"""
Log utils. Used to manage the output format of AutoGL
"""

import logging


def get_logger(name):
    """
    Get the logger of given name

    Parameters
    ----------
    name: str
        The name of logger

    Returns
    -------
    logger: Logger
        The logger generated
    """
    return logging.getLogger(name)



import os
import torch
import logging

import torch.nn.functional as F
import torch.utils.data

import numpy as np
import typing as _typing
import torch.multiprocessing as mp

from typing import Union, Tuple, Sequence, Type, Callable

from tqdm import trange
from copy import deepcopy

from .base import BaseContrastiveTrainer

from .. import register_trainer
from ..base import EarlyStopping, _DummyModel
from ..evaluation import Evaluation, get_feval, Acc
from ...model import (
    BaseAutoModel,
    BaseEncoderMaintainer, 
    BaseDecoderMaintainer,
    DecoderUniversalRegistry
)
from ....datasets import utils
from ....utils import get_logger

from torch_geometric.nn.glob import (
    global_add_pool, global_max_pool, global_mean_pool
)

LOGGER = get_logger("graphcl trainer")

@register_trainer("GraphCLSemisupervisedTrainer")
class GraphCLSemisupervisedTrainer(BaseContrastiveTrainer):
    def __init__(
        self, 
        model: Tuple[BaseEncoderMaintainer, BaseDecoderMaintainer] = None,
        prediction_head: Union[BaseDecoderMaintainer, str, None] = None,
        num_features: Union[int, None] = None,
        num_classes: Union[int,None] = None,
        num_graph_features: Union[int, None] = 0,
        device: Union[torch.device, str] = "auto",
        feval: Union[
            Sequence[str], Sequence[Type[Evaluation]]
        ] = (Acc,),
        views_fn: Union[
            Sequence[str], Sequence[Callable], None
        ] = None,
        aug_ratio: Union[float, Sequence[float]] = 0.2,
        z_dim: Union[int, None] = 128,
        tau: int = 0.5,
        model_path: Union[str, None] = "./models",
        num_workers: int = 0,
        batch_size: int = 128,
        init: bool = False,
        *args,
        **kwargs,
    ):
        """
        This trainer implements the semi-supervised task training method of GraphCL for graph-level tasks with pretraining and finetuning.
        GraphCL is a contrastive method proposed in the paper `Graph Contrastive Learning with Augmentations` in nips 2020.
        <https://proceedings.neurips.cc/paper/2020/hash/3fe230348e9a12c13120749e3f9fa4cd-Abstract.html>
        
        Parameters
        ----------
        model:
            Models can be `str`, `autogl.module.model.encoders.BaseEncoderMaintainer` or a tuple of (encoder, decoder) 
            if need to specify both encoder and decoder. Encoder can be `str` or
            `autogl.module.model.encoders.BaseEncoderMaintainer`, and decoder can be `str`
            or `autogl.module.model.decoders.BaseDecoderMaintainer`.
            If only encoder is specified, decoder will be default to "sumpoolmlp"
        prediction_model: `BaseAutoModel`, `str` or None
            A model used to finetuning
            Only required if `node_level` = True.
        num_features: `int` or None
            The number of features in dataset.
        num_classes: `int` or None
            The number of classes in dataset. Only required when doing semi-supervised tasks.
        num_graph_features: `int` or None
            The number of graph level features in dataset.
        device: `torch.device` or `str`
            The device this trainer will use.
            When `device` = "auto", if GPU exists in the device and dependency is installed, 
            the trainer will give priority to GPU, otherwise CPU will be used
        feval: a sequence of `str` or a sequence of `Evaluation`
            The evaluation methods.
        loss: `str` or `Callable`
            The loss function or the learning objective of contrastive model.
        f_loss: `str` or `Callable`
            The loss function during finetuning stage.
        views_fn: a list of `str`, a list of `Callable` or None
            List of functions or augmentation methods to generate views from give graphs.
            In GraphCL, the number of views_fn should be 2.
            If the element of `views_fn` is str, the options should be "dropN", "permE", "subgraph", "maskN", "random2", "random3", "random4" or None.
        aug_ratio: a list of `float` or a `float`
            The ratio of augmentations, a number between [0, 1).
            If aug_ratio is set as a float, the value will shared by all views_fn.
            If aug_ratio is set as a list of float, the value of this list and views_fn one to one correspondence.
        z_dim: `int`
            The dimension of graph-level representations
        tau: `int`
            The temperature parameter in NT_Xent loss. Only used when `loss` = "NT_Xent"
        model_path: `str` or None
            The directory to restore the saved model.
            If `model_path` = None, the model will not be saved.
        num_workers: `int`
            Number of workers.
        batch_size: `int`
            Batch size for pretraining and inference.
        p_optim: `str` or `torch.optim.Optimizer`
            Optimizer for pretraining.
        p_lr: `float`
            Pretraining learning rate.
        p_lr_scheduler_type: `str`
            Scheduler type for pretraining learning rate.
        p_weight_decay: `float`
            Pretraining weight decay rate.
        p_epoch: `int`
            Pretraining epochs number.
        p_early_stopping: `int`
            Pretraining early stopping round.
        f_optim: `str` or `torch.optim.Optimizer`
            Optimizer for finetuning.
        f_lr: `float`
            Finetuning learning rate.
        f_lr_scheduler_type: `str`
            Scheduler type for finetuning learning rate.
        f_weight_decay: `float`
            Finetuning weight decay rate.
        f_epoch: `int`
            Finetuning epochs number.
        f_early_stopping: `int`
            Finetuning early stopping round.
        init: `bool`
            Whether to initialize the model.
        """
        # set encoder and decoder
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            raise ValueError("The GraphCL trainer must need an encoder and a decoder, so `model` shouldn't be an instance of `BaseAutoModel`")
        else:
            encoder, decoder = model, "sumpoolmlp"
        # init contrastive learning
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            decoder_node=None,
            num_features=num_features,
            num_graph_features=num_graph_features,
            views_fn=views_fn,
            aug_ratio=aug_ratio,
            graph_level=True,
            node_level=False,
            device=device,
            feval=feval,
            z_dim=z_dim,
            z_node_dim=None,
            tau=tau,
            model_path=model_path,
            *args,
            **kwargs
        )
        self.views_fn = views_fn
        self.aug_ratio = aug_ratio
        self._prediction_head = None
        self.num_classes = num_classes
        self.prediction_head = prediction_head
        # recording the information about dataloader
        self.batch_size = batch_size
        self.num_workers = num_workers
        if self.num_workers > 0:
            mp.set_start_method("fork", force=True)

        # hyper parameter space
        self.hyper_parameter_space = [
            {
                "parameterName": "batch_size",
                "type": "INTEGER",
                "maxValue": 128,
                "minValue": 32,
                "scalingType": "LOG",
            },
            {
                "parameterName": "p_epoch",
                "type": "INTEGER",
                "maxValue": 150,
                "minValue": 50,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "p_early_stopping_round",
                "type": "INTEGER",
                "maxValue": 30,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "p_lr",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "p_weight_decay",
                "type": "DOUBLE",
                "maxValue": 5e-3,
                "minValue": 5e-8,
                "scalingType": "LOG"
            },
            {
                "parameterName": "f_epoch",
                "type": "INTEGER",
                "maxValue": 150,
                "minValue": 50,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "f_early_stopping_round",
                "type": "INTEGER",
                "maxValue": 100,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "f_lr",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "f_weight_decay",
                "type": "DOUBLE",
                "maxValue": 5e-3,
                "minValue": 0,
                "scalingType": "LOG"
            },
        ]
        self.hyper_parameters = {
            "batch_size": self.batch_size,
            "p_epoch": self.p_epoch,
            "p_early_stopping_round": self.p_early_stopping_round,
            "p_lr": self.p_lr,
            "p_weight_decay": self.p_weight_decay,
            "f_epoch": self.f_epoch,
            "f_early_stopping_round": self.f_early_stopping_round,
            "f_lr": self.f_lr,
            "f_weight_decay": self.f_weight_decay,
        }
        self.args = args
        self.kwargs = kwargs
        if init:
            self.initialize()

    def _get_views_fn(self, views_fn, aug_ratio):
        # GraphCL only need two kinds of augmentation methods
        assert (views_fn is None) or (len(views_fn) == 2)
        return super()._get_views_fn(views_fn, aug_ratio)

    def _initialize(self):
        self.encoder.initialize()
        if self.decoder is not None:
            self.decoder.initialize(self.encoder)
        if self.prediction_head is not None:
            self.prediction_head.initialize(self.encoder)

    @classmethod
    def get_task_name(cls):
        return "GraphCLSemisupervisedTrainer"

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "p_optimizer": self.p_optimizer,
                "p_learning_rate": self.p_lr,
                "p_max_epoch": self.p_epoch,
                "p_early_stopping_round": self.p_early_stopping_round,
                "f_optimizer": self.f_optimizer,
                "f_learning_rate": self.f_lr,
                "f_max_epoch": self.f_epoch,
                "f_early_stopping_round": self.f_early_stopping_round,
                "encoder": repr(self.encoder),
                "decoder": repr(self.decoder),
                "prediction_head": repr(self.prediction_head)
            }
        )
    
    def _train_finetuning_only(self, dataset):
        """
        Finetuning stage.
        In this stage, we actually trained a combination model with the encoder trained
        during pretraining stage and another projection head with downstream task.
        """
        model = self._compose_model()
        fine_train_loader = utils.graph_get_split(dataset, "train", batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)
        try:
            fine_val_loader = utils.graph_get_split(dataset, "val", batch_size=self.batch_size, num_workers=self.num_workers)
        except ValueError:
            fine_val_loader = utils.graph_get_split(dataset, "train", batch_size=self.batch_size, num_workers=self.num_workers)
        except AttributeError:
            fine_val_loader = utils.graph_get_split(dataset, "train", batch_size=self.batch_size, num_workers=self.num_workers)
        optimizer = self.f_optimizer(model.parameters(), lr=self.f_lr, weight_decay=self.f_weight_decay)
        scheduler = self._get_scheduler("finetune", optimizer)
        with trange(self.f_epoch) as t:
            epoch_loss = 0.0
            for epoch in t:
                t.set_description(f"Finetuning epoch {epoch + 1}")
                model.train()
                for data in fine_train_loader:
                    optimizer.zero_grad()
                    data = data.to(self.device)
                    out = model(data)
                    loss = self.f_loss(out, data.y)
                    loss.backward()
                    optimizer.step()
                    if self.f_lr_scheduler_type:
                        scheduler.step()
                    epoch_loss += loss.item()
                if fine_val_loader is not None:
                    eval_func = (
                        self.feval if not isinstance(self.feval, list) else self.feval[0]
                    )
                    val_loss = self._evaluate(fine_val_loader, eval_func)

                    if eval_func.is_higher_better():
                        val_loss = -val_loss
                    self.f_early_stopping(val_loss, model)
                    if self.f_early_stopping.early_stop:
                        LOGGER.debug("Early stopping at", epoch)
                        break
                else:
                    self.f_early_stopping(epoch_loss, model)
                    if self.f_early_stopping.early_stop:
                        LOGGER.debug("Early stop at", epoch)
                        break
        self.f_early_stopping.load_checkpoint(model)

    def _train_pretraining_only(self, dataset, per_epoch=False):
        for epoch in enumerate(super()._train_pretraining_only(dataset, per_epoch=per_epoch)):
            pass

    def _train_only(self, dataset):
        """
        Training the model on the given dataset.
        """
        self.encoder.encoder.to(self.device)
        self.decoder.decoder.to(self.device)
        self._train_pretraining_only(dataset)
        self._train_finetuning_only(dataset)

    def _predict_only(self, loader, return_label=False):
        model = self._compose_model()
        model.eval()
        pred = []
        label = []
        for data in loader:
            data = data.to(self.device)
            out = model(data)
            pred.append(out)
            label.append(data.y)
        ret = torch.cat(pred, 0)
        label = torch.cat(label, 0)
        if return_label:
            return ret, label
        else:
            return ret

    def duplicate_from_hyper_parameter(self, hp, encoder="same", decoder="same", prediction_head="same", restricted=True):
        """
        The function of duplicating a new instance from the given hyperparameter.

        Parameters
        ----------
        hp: `dict`
            The hyperparameter used in the new instance. Should contain 4 keys "trainer", "encoder"
            "decoder" and "phead", with corresponding hyperparameters as values.
        
        encoder: `str` or `autogl.module.model.encoders.BaseEncoderMaintainer`
            The new encoder
        
        decoder: `str` or `autogl.module.model.decoders.BaseDecoderMaintainer`
            The new decoder

        prediction_head: `str` or `autogl.module.model.decoders.BaseDecoderMaintainer`
            The new predictor

        restricted: `bool`
            If False(True), the hyperparameter should (not) be updated from origin hyperparameter.

        Returns
        -------
        self: `autogl.train.ssl.GraphCLSemisupervisedTrainer`
            A new instance of trainer.

        """
        hp_trainer = hp.get("trainer", {})
        hp_encoder = hp.get("encoder", {})
        hp_decoder = hp.get("decoder", {})
        hp_phead = hp.get("prediction_head", {})
        if not restricted:
            origin_hp = deepcopy(self.hyper_parameters)
            origin_hp.update(hp_trainer)
            hp = origin_hp
        else:
            hp = hp_trainer
        encoder = encoder if encoder != "same" else self.encoder
        decoder = decoder if decoder != "same" else self.decoder
        prediction_head = prediction_head if prediction_head != "same" else self.prediction_head
        encoder = encoder.from_hyper_parameter(hp_encoder)
        decoder.output_dimension = tuple(encoder.get_output_dimensions())[-1]
        if isinstance(encoder, BaseEncoderMaintainer) and isinstance(decoder, BaseDecoderMaintainer):
            decoder = decoder.from_hyper_parameter_and_encoder(hp_decoder, encoder)
        if isinstance(encoder, BaseEncoderMaintainer) and isinstance(prediction_head, BaseDecoderMaintainer):
            prediction_head = prediction_head.from_hyper_parameter_and_encoder(hp_phead, encoder)
        ret = self.__class__(
            model=(encoder, decoder),
            prediction_head=prediction_head,
            num_features=self.num_features,
            num_classes=self.num_classes,
            num_graph_features=self.num_graph_features,
            device=self.device,
            feval=self.feval,
            loss=self.loss,
            f_loss=self.f_loss,
            views_fn=self.views_fn_opt,
            aug_ratio=self.aug_ratio,
            z_dim=self.last_dim,
            tau=self.tau,
            model_path=self.model_path,
            num_workers=self.num_workers,
            batch_size=hp["batch_size"],
            p_optim=self.p_opt_received,
            p_lr=hp["p_lr"],
            p_lr_scheduler_type=self.p_lr_scheduler_type,
            p_epoch=hp["p_epoch"],
            p_early_stopping_round=hp["p_early_stopping_round"],
            p_weight_decay=hp["p_weight_decay"],
            f_optim=self.f_opt_received,
            f_lr=hp["f_lr"],
            f_lr_scheduler_type=self.f_lr_scheduler_type,
            f_epoch=hp["f_epoch"],
            f_early_stopping_round=hp["f_early_stopping_round"],
            f_weight_decay=hp["f_weight_decay"],
            init=True,
            *self.args,
            **self.kwargs
        )

        return ret


@register_trainer("GraphCLUnsupervisedTrainer")
class GraphCLUnsupervisedTrainer(BaseContrastiveTrainer):
    def __init__(
        self, 
        model: Tuple[BaseEncoderMaintainer, BaseDecoderMaintainer] = None,
        prediction_head: Union[BaseDecoderMaintainer, str, None] = None,
        num_features: Union[int, None] = None,
        num_classes: Union[int,None] = None,
        num_graph_features: Union[int, None] = 0,
        device: Union[torch.device, str] = "auto",
        feval: Union[
            Sequence[str], Sequence[Type[Evaluation]]
        ] = (Acc,),
        views_fn: Union[
            Sequence[str], Sequence[Callable], None
        ] = None,
        aug_ratio: Union[float, Sequence[float]] = 0.2,
        z_dim: Union[int, None] = 128,
        num_workers: int = 0,
        batch_size: int = 128,
        eval_interval: int = 10,
        init: bool = False,
        *args,
        **kwargs,
    ):
        """
        This trainer implements the semi-supervised task training method of GraphCL for graph-level tasks with pretraining and finetuning.
        GraphCL is a contrastive method proposed in the paper `Graph Contrastive Learning with Augmentations` in nips 2020.
        <https://proceedings.neurips.cc/paper/2020/hash/3fe230348e9a12c13120749e3f9fa4cd-Abstract.html>
        
        Parameters
        ----------
        model:
            Models can be `str`, `autogl.module.model.encoders.BaseEncoderMaintainer` or a tuple of (encoder, decoder) 
            if need to specify both encoder and decoder. Encoder can be `str` or
            `autogl.module.model.encoders.BaseEncoderMaintainer`, and decoder can be `str`
            or `autogl.module.model.decoders.BaseDecoderMaintainer`.
            If only encoder is specified, decoder will be default to "sumpoolmlp"
        prediction_model: `BaseAutoModel`, `str` or None
            A model used to finetuning
            Only required if `node_level` = True.
        num_features: `int` or None
            The number of features in dataset.
        num_classes: `int` or None
            The number of classes in dataset. Only required when doing semi-supervised tasks.
        num_graph_features: `int` or None
            The number of graph level features in dataset.
        device: `torch.device` or `str`
            The device this trainer will use.
            When `device` = "auto", if GPU exists in the device and dependency is installed, 
            the trainer will give priority to GPU, otherwise CPU will be used
        feval: a sequence of `str` or a sequence of `Evaluation`
            The evaluation methods.
        loss: `str` or `Callable`
            The loss function or the learning objective of contrastive model.
        f_loss: `str` or `Callable`
            The loss function during finetuning stage.
        views_fn: a list of `str`, a list of `Callable` or None
            List of functions or augmentation methods to generate views from give graphs.
            In GraphCL, the number of views_fn should be 2.
            If the element of `views_fn` is str, the options should be "dropN", "permE", "subgraph", "maskN", "random2", "random3", "random4" or None.
        aug_ratio: a list of `float` or a `float`
            The ratio of augmentations, a number between [0, 1).
            If aug_ratio is set as a float, the value will shared by all views_fn.
            If aug_ratio is set as a list of float, the value of this list and views_fn one to one correspondence.
        z_dim: `int`
            The dimension of graph-level representations
        tau: `int`
            The temperature parameter in NT_Xent loss. Only used when `loss` = "NT_Xent"
        model_path: `str` or None
            The directory to restore the saved model.
            If `model_path` = None, the model will not be saved.
        num_workers: `int`
            Number of workers.
        batch_size: `int`
            Batch size for pretraining and inference.
        p_optim: `str` or `torch.optim.Optimizer`
            Optimizer for pretraining.
        p_lr: `float`
            Pretraining learning rate.
        p_lr_scheduler_type: `str`
            Scheduler type for pretraining learning rate.
        p_weight_decay: `float`
            Pretraining weight decay rate.
        p_epoch: `int`
            Pretraining epochs number.
        p_early_stopping: `int`
            Pretraining early stopping round.
        f_optim: `str` or `torch.optim.Optimizer`
            Optimizer for finetuning.
        f_lr: `float`
            Finetuning learning rate.
        f_lr_scheduler_type: `str`
            Scheduler type for finetuning learning rate.
        f_weight_decay: `float`
            Finetuning weight decay rate.
        f_epoch: `int`
            Finetuning epochs number.
        f_early_stopping: `int`
            Finetuning early stopping round.
        init: `bool`
            Whether to initialize the model.
        """
        # set encoder and decoder
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            raise ValueError("The GraphCL trainer must need an encoder and a decoder, so `model` shouldn't be an instance of `BaseAutoModel`")
        else:
            encoder, decoder = model, "sumpoolmlp"
        self.eval_interval = eval_interval
        # init contrastive learning
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            decoder_node=None,
            num_features=num_features,
            num_graph_features=num_graph_features,
            views_fn=views_fn,
            graph_level=True,
            node_level=False,
            device=device,
            feval=feval,
            z_dim=z_dim,
            z_node_dim=None,
            *args,
            **kwargs
        )
        self.views_fn = views_fn
        self.aug_ratio = aug_ratio
        self._prediction_head = None
        self.num_classes = num_classes
        self.prediction_head = prediction_head
        # recording the information about dataloader
        self.batch_size = batch_size
        self.num_workers = num_workers
        if self.num_workers > 0:
            mp.set_start_method("fork", force=True)

        # hyper parameter space
        self.hyper_parameter_space = [
            {
                "parameterName": "batch_size",
                "type": "INTEGER",
                "maxValue": 128,
                "minValue": 32,
                "scalingType": "LOG",
            },
            {
                "parameterName": "p_epoch",
                "type": "INTEGER",
                "maxValue": 150,
                "minValue": 50,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "p_early_stopping_round",
                "type": "INTEGER",
                "maxValue": 30,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "p_lr",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "p_weight_decay",
                "type": "DOUBLE",
                "maxValue": 5e-3,
                "minValue": 5e-8,
                "scalingType": "LOG"
            },
            {
                "parameterName": "f_epoch",
                "type": "INTEGER",
                "maxValue": 150,
                "minValue": 50,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "f_early_stopping_round",
                "type": "INTEGER",
                "maxValue": 100,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "f_lr",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "f_weight_decay",
                "type": "DOUBLE",
                "maxValue": 5e-3,
                "minValue": 0,
                "scalingType": "LOG"
            },
        ]
        self.hyper_parameters = {
            "batch_size": self.batch_size,
            "p_epoch": self.p_epoch,
            "p_early_stopping_round": self.p_early_stopping_round,
            "p_lr": self.p_lr,
            "p_weight_decay": self.p_weight_decay,
            "f_epoch": self.f_epoch,
            "f_early_stopping_round": self.f_early_stopping_round,
            "f_lr": self.f_lr,
            "f_weight_decay": self.f_weight_decay,
        }
        self.args = args
        self.kwargs = kwargs
        if init:
            self.initialize()

    def _get_views_fn(self, views_fn, aug_ratio):
        # GraphCL only need two kinds of augmentation methods
        assert (views_fn is None) or (len(views_fn) == 2)
        return super()._get_views_fn(views_fn, aug_ratio)

    def _initialize(self):
        self.encoder.initialize()
        if self.decoder is not None:
            self.decoder.initialize(self.encoder)
        if self.prediction_head is not None and isinstance(self.prediction_head, BaseDecoderMaintainer):
            self.prediction_head.initialize(self.encoder)

    @classmethod
    def get_task_name(cls):
        return "GraphCLUnsupervisedTrainer"

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "p_optimizer": self.p_optimizer,
                "p_learning_rate": self.p_lr,
                "p_max_epoch": self.p_epoch,
                "p_early_stopping_round": self.p_early_stopping_round,
                "f_optimizer": self.f_optimizer,
                "f_learning_rate": self.f_lr,
                "f_max_epoch": self.f_epoch,
                "f_early_stopping_round": self.f_early_stopping_round,
                "encoder": repr(self.encoder),
                "decoder": repr(self.decoder),
                "prediction_head": repr(self.prediction_head)
            }
        )

    def _get_eval_embed(self, loader):
        self.encoder.encoder.eval()
        num_layers = self.encoder.hyper_parameters['num_layers']
        ret, y = [], []
        for _ in range(num_layers):
            ret.append([])
        with torch.no_grad():
            for data in loader:
                y.append(data.y.numpy())
                data.to(self.device)
                embed = self.encoder.encoder(data)
                for i in range(num_layers):
                    embed[i] = global_add_pool(embed[i], data.batch)
                    ret[i].append(embed[i].cpu().numpy())
        for i in range(num_layers):
            ret[i] = np.concatenate(ret[i], 0)
        y = np.concatenate(y, 0)
        return ret, y

    def _train_only(self, dataset):
        """
        Training the model on the given dataset.
        """
        self.encoder.encoder.to(self.device)
        self.decoder.decoder.to(self.device)
        test_scores = []
        for i, epoch in enumerate(super()._train_pretraining_only(dataset, per_epoch=True)):
            if (i + 1) % self.eval_interval == 0:
                train_loader = utils.graph_get_split(dataset, "train", batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)
                test_loader = utils.graph_get_split(dataset, "test", batch_size=self.batch_size, num_workers=self.num_workers)
                # test_embed, test_lbls = self._get_eval_embed(test_loader)
                if not isinstance(self.prediction_head, BaseDecoderMaintainer):
                    train_embed, train_lbls = self._get_eval_embed(train_loader)
                    self.prediction_head.initialize()
                    self.prediction_head.fit(train_embed, train_lbls)
                    acc = self._evaluate(test_loader)[0]
                    test_scores.append(acc)
                    self.prediction_head.save_checkpoint(acc)
                    self.p_early_stopping(-acc, self.encoder.encoder)
                else:
                    self.encoder.encoder.eval()
                    self.tmp_f_early_stopping = EarlyStopping(
                        patience=self.f_early_stopping_round, verbose=False
                    )
                    self.prediction_head.initialize(self.encoder)
                    model = self.prediction_head.decoder
                    optimizer = self.f_optimizer(model.parameters(), lr=self.f_lr, weight_decay=self.f_weight_decay)
                    scheduler = self._get_scheduler("finetune", optimizer)
                    with trange(self.f_epoch) as t:
                        for epoch in t:
                            model.train()
                            for data in train_loader:
                                optimizer.zero_grad()
                                data = data.to(self.device)
                                embeds = self.encoder.encoder(data)
                                out = model(embeds, data)
                                loss = self.f_loss(out, data.y)
                                loss.backward()
                                optimizer.step()
                                if self.f_lr_scheduler_type:
                                    scheduler.step()
                            eval_func = (
                                self.feval if not isinstance(self.feval, list) else self.feval[0]
                            )
                            val_loss = self._evaluate(test_loader, eval_func)

                            if eval_func.is_higher_better():
                                val_loss = -val_loss
                            self.tmp_f_early_stopping(val_loss, model)
                            if self.f_early_stopping.early_stop:
                                LOGGER.debug("Early stopping at", epoch)
                                break
                                
                    self.tmp_f_early_stopping.load_checkpoint(model)
                    acc = self._evaluate(test_loader)
                    test_scores.append(acc[0])
                    self.f_early_stopping(self.tmp_f_early_stopping.val_loss_min, self.prediction_head.decoder)

        idx = np.argmax(test_scores)
        acc = test_scores[idx]
        if isinstance(self.prediction_head, BaseDecoderMaintainer):
            self.f_early_stopping.load_checkpoint(self.prediction_head.decoder)
        else:
            self.prediction_head.load_checkpoint()
            self.p_early_stopping.load_checkpoint(self.encoder.encoder)
        print("Best epoch %d: acc %.4f" % ((idx + 1) * self.eval_interval, acc))

    def _predict_only(self, loader, return_label=False):
        if isinstance(self.prediction_head, BaseDecoderMaintainer):
            model = self._compose_model()
            model.eval()
            pred = []
            label = []
            for data in loader:
                data = data.to(self.device)
                out = model(data)
                pred.append(out)
                label.append(data.y)
            ret = torch.cat(pred, 0)
            label = torch.cat(label, 0)
        else:
            embeds, label = self._get_eval_embed(loader)
            model = self.prediction_head
            tmpret = model.predict(embeds)
            ret = torch.zeros(tmpret.shape[0], self.num_classes)
            for i in range(tmpret.shape[0]):
                ret[i, tmpret[i]] = 1.
        if return_label:
            return ret, label
        else:
            return ret

    def duplicate_from_hyper_parameter(self, hp, encoder="same", decoder="same", prediction_head="same", restricted=True):
        """
        The function of duplicating a new instance from the given hyperparameter.

        Parameters
        ----------
        hp: `dict`
            The hyperparameter used in the new instance. Should contain 4 keys "trainer", "encoder"
            "decoder" and "phead", with corresponding hyperparameters as values.
        
        encoder: The new encoder
            Encoder can be `str` or `autogl.module.model.encoders.BaseEncoderMaintainer`
        
        decoder: The new decoder
            Decoder can be `str` or `autogl.module.model.decoders.BaseDecoderMaintainer`

        prediction_head: The new prediction_head
            Prediction head can be `str` or `autogl.module.model.decoders.BaseDecoderMaintainer`

        restricted: `bool`
            If False(True), the hyperparameter should (not) be updated from origin hyperparameter.

        Returns
        -------
        self: `autogl.train.ssl.GraphCLSemisupervisedTrainer`
            A new instance of trainer.

        """
        hp_trainer = hp.get("trainer", {})
        hp_encoder = hp.get("encoder", {})
        hp_decoder = hp.get("decoder", {})
        hp_phead = hp.get("prediction_head", {})
        if not restricted:
            origin_hp = deepcopy(self.hyper_parameters)
            origin_hp.update(hp_trainer)
            hp = origin_hp
        else:
            hp = hp_trainer
        encoder = encoder if encoder != "same" else self.encoder
        decoder = decoder if decoder != "same" else self.decoder
        prediction_head = prediction_head if prediction_head != "same" else self.prediction_head
        encoder = encoder.from_hyper_parameter(hp_encoder)
        decoder.output_dimension = tuple(encoder.get_output_dimensions())[-1]
        if isinstance(encoder, BaseEncoderMaintainer) and isinstance(decoder, BaseDecoderMaintainer):
            decoder = decoder.from_hyper_parameter_and_encoder(hp_decoder, encoder)
        if isinstance(encoder, BaseEncoderMaintainer) and isinstance(prediction_head, BaseDecoderMaintainer):
            prediction_head = prediction_head.from_hyper_parameter_and_encoder(hp_phead, encoder)
        ret = self.__class__(
            model=(encoder, decoder),
            prediction_head=prediction_head,
            num_features=self.num_features,
            num_classes=self.num_classes,
            num_graph_features=self.num_graph_features,
            device=self.device,
            feval=self.feval,
            loss=self.loss,
            f_loss=self.f_loss,
            views_fn=self.views_fn_opt,
            aug_ratio=self.aug_ratio,
            z_dim=self.last_dim,
            tau=self.tau,
            model_path=self.model_path,
            num_workers=self.num_workers,
            batch_size=hp["batch_size"],
            eval_interval=self.eval_interval,
            p_optim=self.p_opt_received,
            p_lr=hp["p_lr"],
            p_lr_scheduler_type=self.p_lr_scheduler_type,
            p_epoch=hp["p_epoch"],
            p_early_stopping_round=hp["p_early_stopping_round"],
            p_weight_decay=hp["p_weight_decay"],
            f_optim=self.f_opt_received,
            f_lr=hp["f_lr"],
            f_lr_scheduler_type=self.f_lr_scheduler_type,
            f_epoch=hp["f_epoch"],
            f_early_stopping_round=hp["f_early_stopping_round"],
            f_weight_decay=hp["f_weight_decay"],
            init=True,
            *self.args,
            **self.kwargs
        )

        return ret


import logging
import dgl
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl.function as fn
from dgl.nn.functional import edge_softmax

from .. import register_model
from .base import BaseHeteroModelMaintainer
from ..base import activate_func
from .....utils import get_logger

LOGGER = get_logger("HGTModel")

def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args


class HGTLayer(nn.Module):
    def __init__(self,
                 in_dim,
                 out_dim,
                 node_dict,
                 edge_dict,
                 n_heads,
                 dropout = 0.2,
                 use_norm = False,
                 out_key = None):
        super(HGTLayer, self).__init__()

        self.in_dim        = in_dim
        self.out_dim       = out_dim
        self.node_dict     = node_dict
        self.edge_dict     = edge_dict
        self.out_key = out_key
        self.num_types     = len(node_dict)
        self.num_relations = len(edge_dict)
        self.total_rel     = self.num_types * self.num_relations * self.num_types
        self.n_heads       = n_heads
        self.d_k           = out_dim // n_heads
        self.sqrt_dk       = math.sqrt(self.d_k)
        self.att           = None

        self.k_linears   = nn.ModuleList()
        self.q_linears   = nn.ModuleList()
        self.v_linears   = nn.ModuleList()
        self.a_linears   = nn.ModuleList()
        self.norms       = nn.ModuleList()
        self.use_norm    = use_norm

        for t in range(self.num_types):
            self.k_linears.append(nn.Linear(in_dim,   out_dim))
            self.q_linears.append(nn.Linear(in_dim,   out_dim))
            self.v_linears.append(nn.Linear(in_dim,   out_dim))
            self.a_linears.append(nn.Linear(out_dim,  out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))

        self.relation_pri   = nn.Parameter(torch.ones(self.num_relations, self.n_heads))
        self.relation_att   = nn.Parameter(torch.Tensor(self.num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg   = nn.Parameter(torch.Tensor(self.num_relations, n_heads, self.d_k, self.d_k))
        self.skip           = nn.Parameter(torch.ones(self.num_types))
        self.drop           = nn.Dropout(dropout)

        nn.init.xavier_uniform_(self.relation_att)
        nn.init.xavier_uniform_(self.relation_msg)

    def forward(self, G, h):
        with G.local_scope():
            node_dict, edge_dict = self.node_dict, self.edge_dict
            for srctype, etype, dsttype in G.canonical_etypes:
                sub_graph = G[srctype, etype, dsttype]

                k_linear = self.k_linears[node_dict[srctype]]
                v_linear = self.v_linears[node_dict[srctype]]
                q_linear = self.q_linears[node_dict[dsttype]]

                k = k_linear(h[srctype]).view(-1, self.n_heads, self.d_k)
                v = v_linear(h[srctype]).view(-1, self.n_heads, self.d_k)
                q = q_linear(h[dsttype]).view(-1, self.n_heads, self.d_k)

                e_id = self.edge_dict[etype]

                relation_att = self.relation_att[e_id]
                relation_pri = self.relation_pri[e_id]
                relation_msg = self.relation_msg[e_id]

                k = torch.einsum("bij,ijk->bik", k, relation_att)
                v = torch.einsum("bij,ijk->bik", v, relation_msg)

                sub_graph.srcdata['k'] = k
                sub_graph.dstdata['q'] = q
                sub_graph.srcdata['v_%d' % e_id] = v

                sub_graph.apply_edges(fn.v_dot_u('q', 'k', 't'))
                attn_score = sub_graph.edata.pop('t').sum(-1) * relation_pri / self.sqrt_dk
                attn_score = edge_softmax(sub_graph, attn_score, norm_by='dst')

                sub_graph.edata['t'] = attn_score.unsqueeze(-1)

            G.multi_update_all({etype : (fn.u_mul_e('v_%d' % e_id, 't', 'm'), fn.sum('m', 't')) \
                                for etype, e_id in edge_dict.items()}, cross_reducer = 'mean')

            new_h = {}
            for ntype in G.ntypes:
                '''
                    Step 3: Target-specific Aggregation
                    x = norm( W[node_type] * gelu( Agg(x) ) + x )
                '''
                n_id = node_dict[ntype]
                alpha = torch.sigmoid(self.skip[n_id])
                t = G.nodes[ntype].data['t'].view(-1, self.out_dim)
                trans_out = self.drop(self.a_linears[n_id](t))
                trans_out = trans_out * alpha + h[ntype] * (1-alpha)
                if self.use_norm:
                    new_h[ntype] = self.norms[n_id](trans_out)
                else:
                    new_h[ntype] = trans_out
            return new_h

class HGT(nn.Module):
    def __init__(self, args):

        super(HGT, self).__init__()
        self.args = args 
        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_layers",
                    "hidden",
                    "heads",
                    "dropout",
                    "act",
                    "use_norm",
                    "node_dict",
                    "edge_dict",
                    "out_key"
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        self.node_dict = self.args["node_dict"]
        self.edge_dict = self.args["edge_dict"]
        self.out_key = self.args["out_key"]
        self.gcs = nn.ModuleList()
        self.num_layers = int(self.args["num_layers"])

        hidden = self.args["hidden"]*self.args["heads"]

        self.adapt_ws  = nn.ModuleList()
        for t in range(len(self.node_dict)):
            self.adapt_ws.append(nn.Linear(self.args["features_num"], hidden))

        for i in range(self.num_layers):
            self.gcs.append(HGTLayer(hidden, hidden, self.node_dict, self.edge_dict, \
                self.args["heads"], use_norm = self.args["use_norm"], dropout = self.args["dropout"]))
            
        self.out = nn.Linear(hidden, self.args["num_class"])

    def forward(self, G):
        h = {}
        for ntype in G.ntypes:
            n_id = self.node_dict[ntype]
            h[ntype] = activate_func(self.adapt_ws[n_id](G.nodes[ntype].data['feat']), self.args["act"])
        for i in range(self.num_layers):
            h = self.gcs[i](G, h)
        return self.out(h[self.out_key])

@register_model("hgt")
class AutoHGT(BaseHeteroModelMaintainer):
    r"""
    AutoHGT.
    The model used in this automodel is HGT, i.e., the graph convolutional network from the
    `"Heterogeneous Graph Transformer" <https://arxiv.org/abs/2003.01332>`_ paper.
        
    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`.
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    dataset: `autogl.datasets`.
        Hetero Graph Dataset in autogl.

    """
    def __init__(
        self, num_features=None, num_classes=None, device=None, init=False, dataset=None, **args
    ):
        super().__init__(num_features, num_classes, device, dataset, **args)
        
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden", 
                "type": "INTEGER", 
                "minValue": 8, 
                "maxValue": 128,
                "scalingType": "LOG"
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "heads",
                "type": "DISCRETE",
                "feasiblePoints": "2,4,8,16",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh","gelu"],  # add F.gelu()
            },
            {
                "parameterName": "use_norm",
                "type": "CATEGORICAL",
                "feasiblePoints": [True, False], 
            },
        ]

        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": 64,
            "heads": 4,
            "dropout": 0.2,
            "act": "gelu",
            "use_norm": True
        }

        if init is True:
            self.initialize()

    def _initialize(self):
        self._model = HGT(dict(
            features_num=self.input_dimension, 
            num_class=self.output_dimension, 
            out_key=self.out_key,
            node_dict=self.node_dict,
            edge_dict=self.edge_dict,
            **self.hyper_parameters
        )).to(self.device)

    def from_dataset(self, dataset):
        G: dgl.DGLGraph = dataset[0]
        # generate edge and node dict
        self.register_parameter("out_key", dataset.schema["target_node_type"])
        self.register_parameter("node_dict", dict(zip(G.ntypes, range(len(G.ntypes)))))
        self.register_parameter("edge_dict", dict(zip(G.etypes, range(len(G.etypes)))))



# NTXent_loss from <https://github.com/Shen-Lab/GraphCL/>
import torch
import torch.nn as nn

def NTXent_loss(zs, tau=0.5, norm=True):
    batch_size, _ = zs[0].size()
    sim_matrix = torch.einsum('ik,jk->ij', zs[0], zs[1])
    if norm:
        z1_abs = zs[0].norm(dim=1)
        z2_abs = zs[1].norm(dim=1)
        sim_matrix = sim_matrix / torch.einsum('i,j->ij', z1_abs, z2_abs)
    sim_matrix = torch.exp(sim_matrix/tau)
    pos_sim = sim_matrix[range(batch_size), range(batch_size)]
    loss = pos_sim / (sim_matrix.sum(dim=1) - pos_sim)
    loss = - torch.log(loss).mean()
    
    return loss



from .netlsd import SgNetLSD
from .base import BaseGraph
from .nx import (
    register_nx,
    NxGraph,
    nxfunc,
    NxLargeCliqueSize,
    NxAverageClusteringApproximate,
    NxDegreeAssortativityCoefficient,
    NxDegreePearsonCorrelationCoefficient,
    NxHasBridge,
    NxGraphCliqueNumber,
    NxGraphNumberOfCliques,
    NxTransitivity,
    NxAverageClustering,
    NxIsConnected,
    NxNumberConnectedComponents,
    NxIsDistanceRegular,
    NxLocalEfficiency,
    NxGlobalEfficiency,
    NxIsEulerian,
)

__all__ = [
    "SgNetLSD",
    "BaseGraph",
    "register_nx",
    "NxGraph",
    "nxfunc",
    "NxLargeCliqueSize",
    "NxAverageClusteringApproximate",
    "NxDegreeAssortativityCoefficient",
    "NxDegreePearsonCorrelationCoefficient",
    "NxHasBridge",
    "NxGraphCliqueNumber",
    "NxGraphNumberOfCliques",
    "NxTransitivity",
    "NxAverageClustering",
    "NxIsConnected",
    "NxNumberConnectedComponents",
    "NxIsDistanceRegular",
    "NxLocalEfficiency",
    "NxGlobalEfficiency",
    "NxIsEulerian",
]



from ._basic import OneHotFeatureGenerator
from ._eigen import EigenFeatureGenerator
from ._graphlet import GraphletGenerator
from ._page_rank import PageRankFeatureGenerator
from ._pyg import (
    LocalDegreeProfileGenerator,
    NormalizeFeatures,
    OneHotDegreeGenerator
)

__all__ = [
    "OneHotFeatureGenerator",
    "EigenFeatureGenerator",
    "GraphletGenerator",
    "PageRankFeatureGenerator",
    "LocalDegreeProfileGenerator",
    "NormalizeFeatures",
    "OneHotDegreeGenerator"
]



import logging
import torch.nn.functional
import typing as _typing
import dgl
from dgl.nn.pytorch.conv import GATConv
from .. import base_encoder, encoder_registry
from ... import _utils


class GATUtils:
    @classmethod
    def to_total_hidden_dimensions(
            cls, per_head_dimensions: _typing.Sequence[int],
            num_hidden_heads: int, num_output_heads: int, concat_last: bool = False,
    ):
        return [
            d * (num_hidden_heads if layer < (len(per_head_dimensions) - 1) else (num_output_heads if concat_last else 1))
            for layer, d in enumerate(per_head_dimensions)
        ]


class GAT(torch.nn.Module):
    def __init__(
            self, input_dimension: int, dimensions: _typing.Sequence[int],
            num_hidden_heads: int, num_output_heads: int,
            act: _typing.Optional[str], dropout: _typing.Optional[float], concat_last: bool = True
    ):
        super(GAT, self).__init__()
        dimensions = GATUtils.to_total_hidden_dimensions(
            dimensions, num_hidden_heads, num_output_heads, concat_last=concat_last
        )
        num_layers: int = len(dimensions)
        self.__convolutions: torch.nn.ModuleList = torch.nn.ModuleList()
        for layer in range(len(dimensions)):
            num_heads: int = (
                num_hidden_heads
                if layer < num_layers - 1
                else num_output_heads
            )
            if dimensions[layer] % num_heads != 0:
                raise ValueError
            self.__convolutions.append(
                GATConv(
                    dimensions[layer - 1] if layer > 0 else input_dimension,
                    dimensions[layer] // num_heads, num_heads,
                    dropout, dropout
                )
            )
        self.__activation: _typing.Optional[str] = act
        self.__concat_last = concat_last

    def forward(
            self, graph: dgl.DGLGraph, *__args, **__kwargs
    ) -> _typing.Iterable[torch.Tensor]:
        num_layers = len(self.__convolutions)
        x: torch.Tensor = graph.ndata['feat']
        results = [x]
        for layer in range(num_layers):
            if layer < num_layers - 1 or self.__concat_last:
                x = self.__convolutions[layer](graph, x).flatten(1)
            else:
                x = self.__convolutions[layer](graph, x).mean(1)
            if layer < num_layers - 1:
                x = _utils.activation.activation_func(x, self.__activation)
            results.append(x)
        logging.debug("{:d} layer, each layer shape {:s}".format(len(results), " ".join([str(x.shape) for x in results])))
        return results


@encoder_registry.EncoderUniversalRegistry.register_encoder('gat')
@encoder_registry.EncoderUniversalRegistry.register_encoder('gat_encoder')
class GATEncoderMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    r"""
    AutoGAT. The model used in this automodel is GAT, i.e., the graph attentional network from the `"Graph Attention Networks"
    <https://arxiv.org/abs/1710.10903>`_ paper. The layer is

    .. math::
        \mathbf{x}^{\prime}_i = \alpha_{i,i}\mathbf{\Theta}\mathbf{x}_{i} +
        \sum_{j \in \mathcal{N}(i)} \alpha_{i,j}\mathbf{\Theta}\mathbf{x}_{j}
    where the attention coefficients :math:`\alpha_{i,j}` are computed as
    
    .. math::
        \alpha_{i,j} =
        \frac{
        \exp\left(\mathrm{LeakyReLU}\left(\mathbf{a}^{\top}
        [\mathbf{\Theta}\mathbf{x}_i \, \Vert \, \mathbf{\Theta}\mathbf{x}_j]
        \right)\right)}
        {\sum_{k \in \mathcal{N}(i) \cup \{ i \}}
        \exp\left(\mathrm{LeakyReLU}\left(\mathbf{a}^{\top}
        [\mathbf{\Theta}\mathbf{x}_i \, \Vert \, \mathbf{\Theta}\mathbf{x}_k]
        \right)\right)}.
    
    Parameters
    ----------
    input_dimension: `Optional[int]`
        The dimension of input features.
    final_dimension: `Optional[int]`
        The dimension of final features.
    device: `torch.device` or `str` or `int`
        The device where model will be running on.
    kwargs:
        Other parameters.
    """

    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(GATEncoderMaintainer, self).__init__(
            input_dimension, final_dimension, device, *args, **kwargs
        )
        self.hyper_parameters: _typing.Mapping[str, _typing.Any] = {
            "num_layers": 2,
            "hidden": [32],
            "heads": 4,
            "dropout": 0.2,
            "act": "leaky_relu",
        }
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "heads",
                "type": "DISCRETE",
                "feasiblePoints": "2,4,8,16",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

    def get_output_dimensions(self) -> _typing.Iterable[int]:
        temp = list(self.hyper_parameters["hidden"])
        concat_last = True
        if (
                self.final_dimension not in (Ellipsis, None) and
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            temp.append(self.final_dimension)
            concat_last = False
        output_dimensions = [self.input_dimension]
        output_dimensions.extend(
            GATUtils.to_total_hidden_dimensions(
                temp,
                self.hyper_parameters.get("num_hidden_heads", self.hyper_parameters["heads"]),
                self.hyper_parameters.get("num_output_heads", 1), concat_last=concat_last
            )
        )
        return output_dimensions

    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters["hidden"])
        concat_last = True
        if (
                self.final_dimension not in (Ellipsis, None) and
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
            concat_last = False
        logging.debug("current dimensions %s", dimensions)
        self._encoder: torch.nn.Module = GAT(
            self.input_dimension, dimensions,
            self.hyper_parameters.get("num_hidden_heads", self.hyper_parameters["heads"]),
            self.hyper_parameters.get("num_output_heads", 1),
            self.hyper_parameters.get("act"),
            self.hyper_parameters.get("dropout"),
            concat_last
        ).to(self.device)
        return True



from ._netlsd import NetLSD
from ._networkx import (
    NXLargeCliqueSize,
    NXDegreeAssortativityCoefficient,
    NXDegreePearsonCorrelationCoefficient,
    NXHasBridges,
    NXGraphCliqueNumber,
    NXGraphNumberOfCliques,
    NXTransitivity,
    NXAverageClustering,
    NXIsConnected,
    NXNumberConnectedComponents,
    NXIsDistanceRegular,
    NXLocalEfficiency,
    NXGlobalEfficiency,
    NXIsEulerian,
)



import torch
import typing as _typing
import autogl
from autogl.data.graph import GeneralStaticGraph
from .._feature_engineer import FeatureEngineer
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry


class BaseFeatureGenerator(FeatureEngineer):
    def __init__(self, override_features: bool = False):
        super(BaseFeatureGenerator, self).__init__()
        if not isinstance(override_features, bool):
            raise TypeError
        else:
            self._override_features: bool = override_features

    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        raise NotImplementedError

    def __transform_homogeneous_static_graph(
            self, homogeneous_static_graph: GeneralStaticGraph
    ) -> GeneralStaticGraph:
        if not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError("Provided static graph must be homogeneous")
        if 'x' in homogeneous_static_graph.nodes.data:
            feature_key: _typing.Optional[str] = 'x'
            features: _typing.Optional[torch.Tensor] = (
                homogeneous_static_graph.nodes.data['x']
            )
        elif 'feat' in homogeneous_static_graph.nodes.data:
            feature_key: _typing.Optional[str] = 'feat'
            features: _typing.Optional[torch.Tensor] = (
                homogeneous_static_graph.nodes.data['feat']
            )
        else:
            feature_key: _typing.Optional[str] = None
            features: _typing.Optional[torch.Tensor] = None
        if 'y' in homogeneous_static_graph.nodes.data:
            label: _typing.Optional[torch.Tensor] = (
                homogeneous_static_graph.nodes.data['y']
            )
        elif 'label' in homogeneous_static_graph.nodes.data:
            label: _typing.Optional[torch.Tensor] = (
                homogeneous_static_graph.nodes.data['label']
            )
        else:
            label: _typing.Optional[torch.Tensor] = None
        if (
                'edge_weight' in homogeneous_static_graph.edges.data and
                homogeneous_static_graph.edges.data['edge_weight'].dim() == 1
        ):
            edge_weight: torch.Tensor = (
                homogeneous_static_graph.edges.data['edge_weight']
            )
        else:
            edge_weight: torch.Tensor = torch.ones(
                homogeneous_static_graph.edges.connections.size(1)
            )
        data = autogl.data.Data(
            edge_index=homogeneous_static_graph.edges.connections,
            x=features, y=label
        )
        setattr(data, "edge_weight", edge_weight)
        extracted_features: torch.Tensor = self._extract_nodes_feature(data)
        if isinstance(feature_key, str):
            nodes_features: torch.Tensor = (
                homogeneous_static_graph.nodes.data[feature_key].view(-1, 1)
                if homogeneous_static_graph.nodes.data[feature_key].dim() == 1
                else homogeneous_static_graph.nodes.data[feature_key]
            )
            assert extracted_features.size(0) == nodes_features.size(0)
            assert extracted_features.dim() == nodes_features.dim() == 2
            homogeneous_static_graph.nodes.data[feature_key] = (
                extracted_features.to(nodes_features.device)
                if self._override_features
                else torch.cat(
                    [nodes_features, extracted_features.to(nodes_features.device)], dim=-1
                )
            )
        else:
            if autogl.backend.DependentBackend.is_pyg():
                homogeneous_static_graph.nodes.data['x'] = extracted_features
            elif autogl.backend.DependentBackend.is_dgl():
                homogeneous_static_graph.nodes.data['feat'] = extracted_features
        return homogeneous_static_graph

    def _transform(
            self, data: _typing.Union[GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[GeneralStaticGraph, _typing.Any]:
        if isinstance(data, GeneralStaticGraph):
            return self.__transform_homogeneous_static_graph(data)
        else:
            data.x = self._extract_nodes_feature(data)
            return data


@DataPreprocessorUniversalRegistry.register_data_preprocessor("OneHot".lower())
class OneHotFeatureGenerator(BaseFeatureGenerator):
    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        num_nodes: int = (
            data.x.size(0)
            if data.x is not None and isinstance(data.x, torch.Tensor)
            else (data.edge_index.max().item() + 1)
        )
        return torch.eye(num_nodes)



import torch
import typing as _typing
from . import _general_static_graph
from ._general_static_graph_default_implementation import (
    HeterogeneousNodesContainer, HeterogeneousNodesContainerImplementation,
    HomogeneousEdgesContainerImplementation,
    HeterogeneousEdgesAggregation, HeterogeneousEdgesAggregationImplementation,
    StaticGraphDataAggregation, GeneralStaticGraphImplementation
)


class GeneralStaticGraphGenerator:
    @classmethod
    def create_heterogeneous_static_graph(
            cls, heterogeneous_nodes_data: _typing.Mapping[str, _typing.Mapping[str, torch.Tensor]],
            heterogeneous_edges: _typing.Mapping[
                _typing.Tuple[str, str, str],
                _typing.Union[
                    torch.Tensor,
                    _typing.Tuple[
                        torch.Tensor,
                        _typing.Optional[_typing.Mapping[str, torch.Tensor]]
                    ]
                ]
            ],
            graph_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ) -> _general_static_graph.GeneralStaticGraph:
        _heterogeneous_nodes_container: HeterogeneousNodesContainer = (
            HeterogeneousNodesContainerImplementation(heterogeneous_nodes_data)
        )
        _heterogeneous_edges_aggregation: HeterogeneousEdgesAggregation = (
            HeterogeneousEdgesAggregationImplementation()
        )
        for canonical_edge_type, specific_typed_edges in heterogeneous_edges.items():
            if isinstance(specific_typed_edges, torch.Tensor):
                connections = specific_typed_edges
                data = None
            elif (
                    isinstance(specific_typed_edges, _typing.Sequence) and
                    len(specific_typed_edges) == 2 and
                    isinstance(specific_typed_edges[0], torch.Tensor) and
                    (
                            specific_typed_edges[1] is None or
                            isinstance(specific_typed_edges[1], _typing.Mapping)
                    )
            ):
                connections = specific_typed_edges[0]
                data = specific_typed_edges[1]
            else:
                raise TypeError
            _heterogeneous_edges_aggregation[canonical_edge_type] = (
                HomogeneousEdgesContainerImplementation(connections, data)
            )
        return GeneralStaticGraphImplementation(
            _heterogeneous_nodes_container,
            _heterogeneous_edges_aggregation,
            StaticGraphDataAggregation(graph_data)
        )

    @classmethod
    def create_homogeneous_static_graph(
            cls, nodes_data: _typing.Mapping[str, torch.Tensor],
            edges_connections: torch.Tensor,
            edges_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...,
            graph_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ) -> _general_static_graph.GeneralStaticGraph:
        _heterogeneous_nodes_container: HeterogeneousNodesContainer = (
            HeterogeneousNodesContainerImplementation({'': nodes_data})
        )
        _heterogeneous_edges_aggregation: HeterogeneousEdgesAggregation = (
            HeterogeneousEdgesAggregationImplementation()
        )
        _heterogeneous_edges_aggregation[('', '', '')] = (
            HomogeneousEdgesContainerImplementation(edges_connections, edges_data)
        )
        return GeneralStaticGraphImplementation(
            _heterogeneous_nodes_container,
            _heterogeneous_edges_aggregation,
            StaticGraphDataAggregation(graph_data)
        )



"""
Ensemble module.
"""

from collections import Counter
import numpy as np

from .base import BaseEnsembler
from . import register_ensembler
from ...utils import get_logger

VOTE_LOGGER = get_logger("voting")


@register_ensembler("voting")
class Voting(BaseEnsembler):
    """
    An ensembler using the voting method.

    Parameters
    ----------
    ensemble_size : int
        The number of base models selected by the voter. These selected models can be redundant. Default as 10.
    """

    def __init__(self, ensemble_size=10, *args, **kwargs):
        super().__init__()
        self.ensemble_size = ensemble_size

    def fit(self, predictions, label, identifiers, feval, *args, **kwargs):
        """
        Fit the ensembler to the given data using Rich Caruana's ensemble selection method.

        Parameters
        ----------
        predictions : a list of np.ndarray
            Predictions of base learners (corresponding to the elements in identifiers).
        labels : a list of int
            Class labels of instances.
        identifiers : a list of str
            The names of base models.
        feval : (a list of) instances in autogl.module.train.evaluate
            Performance evaluation metrices.

        Returns
        -------
        (a list of) ``float``
            The validation performance of the final voter.
        """

        self._re_initialize(identifiers, len(predictions))

        if not isinstance(feval, list):
            feval = [feval]

        weights = self._specify_weights(predictions, label, feval)
        self.model_to_weight = dict(zip(self.identifiers, weights))

        VOTE_LOGGER.debug(self.identifiers, weights)

        training_score = self._eval(predictions, label, feval)

        return training_score

    def ensemble(self, predictions, identifiers, *args, **kwargs):
        """
        Ensemble the predictions of base models.

        Parameters
        ----------
        predictions : a list of np.ndarray
            Predictions of base learners (corresponding to the elements in identifiers).
        identifiers : a list of str
            The names of base models.

        Returns
        -------
        np.ndarray
            The ensembled predictions.
        """

        weights = np.zeros([len(predictions)])
        for idx, model in enumerate(identifiers):
            weights[idx] = self.model_to_weight[model]
        weights = weights / np.sum(weights)

        return np.average(predictions, axis=0, weights=weights)

    def _specify_weights(self, predictions, label, feval):
        ensemble_prediction = []
        combinations = []
        history = []

        for i in range(self.ensemble_size):
            eval_score_full = []
            eval_score = np.zeros([self.n_models])

            for j, pred in enumerate(predictions):
                ensemble_prediction.append(pred)
                pred_mean = np.mean(ensemble_prediction, axis=0)
                eval_score_full.append(
                    [
                        fx.evaluate(pred_mean, label)
                        * (1 if fx.is_higher_better else -1)
                        for fx in feval
                    ]
                )
                eval_score[j] = eval_score_full[-1][0]
                ensemble_prediction.pop()

            best_model = np.argmax(eval_score)
            ensemble_prediction.append(predictions[best_model])
            history.append(eval_score_full[best_model])
            combinations.append(best_model)

        frequency = Counter(combinations).most_common()
        weights = np.zeros([self.n_models])
        for model, freq in frequency:
            weights[model] = float(freq)

        weights = weights / np.sum(weights)

        return weights

    def _re_initialize(self, identifiers, n_models):
        self.identifiers = identifiers
        self.n_models = n_models

    def _eval(self, predictions, label, feval):
        pred_ensemble = self.ensemble(predictions, self.identifiers)
        return [fx.evaluate(pred_ensemble, label) for fx in feval]




from numba.core.errors import reset_terminal
from numpy.lib.arraysetops import isin
from scipy import sparse
import torch
import torch.nn as nn
from torch.nn import Linear, Sequential, ReLU
import torch.nn.functional as F
from torch.nn import Parameter
from torch_geometric.nn import JumpingKnowledge
from torch_sparse import SparseTensor
import torch_sparse
from torch import Tensor
import numpy as np 
import scipy
import scipy.sparse as sp
from torch_geometric.utils import num_nodes, to_scipy_sparse_matrix,from_scipy_sparse_matrix,remove_self_loops,add_self_loops,add_remaining_self_loops
from numba import njit,jit
from tqdm import tqdm
from scipy.sparse import lil_matrix
from deeprobust.graph.utils import *
from sklearn.metrics import jaccard_score
from sklearn.metrics.pairwise import euclidean_distances, cosine_similarity
from sklearn.preprocessing import normalize

from . import register_nas_space
from .base import BaseSpace
from ...model import BaseAutoModel
from ....utils import get_logger
from .operation import gnn_map,act_map
import typing as _typ
from ..backend import *

from torch_geometric.nn import (
    # GATConv,
    GCNConv,
    ChebConv,
    GatedGraphConv,
    ARMAConv,
    SGConv,
)
from .gasso_space import GATConv,GINConv,SAGEConv

def gnn_map(gnn_name, in_dim, out_dim, concat=False, bias=True) -> nn.Module:
    """

    :param gnn_name:
    :param in_dim:
    :param out_dim:
    :param concat: for gat, concat multi-head output or not
    :return: GNN model
    """
    if gnn_name == "gat_8":
        return GATConv(in_dim, out_dim, 8, concat=concat, bias=bias)
    elif gnn_name == "gat_6":
        return GATConv(in_dim, out_dim, 6, concat=concat, bias=bias)
    elif gnn_name == "gat_4":
        return GATConv(in_dim, out_dim, 4, concat=concat, bias=bias)
    elif gnn_name == "gat_2":
        return GATConv(in_dim, out_dim, 2, concat=concat, bias=bias)
    elif gnn_name in ["gat_1", "gat"]:
        return GATConv(in_dim, out_dim, 1, concat=concat, bias=bias)
    elif gnn_name == "gcn":
        return GCNConv(in_dim, out_dim)
    elif gnn_name == "cheb":
        return ChebConv(in_dim, out_dim, K=2, bias=bias)
    elif gnn_name == "sage":
        return SAGEConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "gated":
        return GatedGraphConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "arma":
        return ARMAConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "sg":
        return SGConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "gin":
        return GINConv(torch.nn.Linear(in_dim, out_dim))
    elif gnn_name == "linear":
        return LinearConv(in_dim, out_dim, bias=bias)
    raise KeyError("Cannot parse key %s" % (gnn_name))


PRE_ROB_OPS={
    'identity': lambda: RobustIdentity(),
    'svd': lambda: SVD(),
    'jaccard': lambda: Jaccard(),
    'gnnguard': lambda: GNNGuard(use_gate=False),
    'vpn': lambda: VPN(),
}

ROB_OPS={
    'identity': lambda: RobustIdentity(),
    'svd': lambda: SVD(),
    'jaccard': lambda: Jaccard(),
    'gnnguard': lambda: GNNGuard(),
    'vpn': lambda: VPN(),
}

# class RobOp(nn.Module):
#     def __init__(self, op, ):
#         super(RobOp, self).__init__()
#         self._op = ROB_OPS[ROB_PRIMITIVES[idx]]()

#     def forward(self, edge_index, edge_weight, features):
#         return self._op(edge_index, edge_weight,  features=features)

class RobustIdentity(nn.Module):
    
    def __init__(self):
        super(RobustIdentity, self).__init__()

    def forward(self, edge_index, edge_weight, features):
        return edge_weight
    
    def check_dense_matrix(self, symmetric=False):
        torch.cuda.empty_cache()
        self.modified_adj= torch.clamp(self.modified_adj,min=0,max=1)
        torch.cuda.empty_cache()

class SVD(RobustIdentity):
    def __init__(self):
        super(SVD, self).__init__()

    def forward(self, edge_index, edge_weight, features,  k=20):
        torch.cuda.empty_cache()
        # print('=== GCN-SVD: rank={} ==='.format(k))
        # if _A.is_sparse:
        # _A = torch.sparse.FloatTensor(edge_index, edge_weight, (features.size(0),features.size(0)))
        # sp_A= torch_sparse_to_scipy_sparse(_A)
        device = edge_index.device
        i = edge_index.cpu().numpy()
        sp_A = sp.csr_matrix((edge_weight.detach().cpu().numpy(), (i[0], i[1])), shape=(features.size(0),features.size(0)))
        row, col = sp_A.tocoo().row, sp_A.tocoo().col
        modified_adj = cal_svd(sp_A, k=k)
        adj_values = torch.tensor(modified_adj[row, col], dtype=torch.float32).to(device)
        adj_values = torch.clamp(adj_values, min=0)
        # else:
        #     sp_A = _A.to_sparse()
        #     modified_adj = cal_svd(torch_sparse_to_scipy_sparse(sp_A),k=k)
        #     modified_adj = torch.from_numpy(modified_adj).to(_A.device)

        #     # sparsify 
        #     self.modified_adj = torch.multiply(sp_A, modified_adj)
        #     self.check_dense_matrix()

        return adj_values

def cal_svd(sp_adj,k):
    adj = sp_adj.asfptype()
    U, S, V = sp.linalg.svds(adj, k=k)
    diag_S = np.diag(S)
    return U @ diag_S @ V



def torch_sparse_to_scipy_sparse(m, return_iv=False):
    '''
    Parameter:
    ----------
    m: torch.sparse matrix
    '''
    i = m.coalesce().indices().detach().cpu().numpy()
    v = m.coalesce().values().detach().cpu().numpy()
    shape = m.coalesce().size()
    
    sp_m = sp.csr_matrix((v, (i[0], i[1])), shape=shape)
    if return_iv:
        return sp_m, i, v
    else:
        return sp_m
        
class Jaccard(RobustIdentity):
    def __init__(self, ):
        super(Jaccard, self).__init__()
    
    def forward(self, edge_index, edge_weight, features, threshold=0.01):
        torch.cuda.empty_cache()
        """Drop dissimilar edges.(Faster version using numba)
        """
        # print("==GCN_Jaccard==")
        self.threshold = threshold
        features = features.detach().cpu().numpy()
        self.binary_feature = (features[features>0]==1).sum()==len(features[features>0])
        # print('Binary feature:',self.binary_feature )
        if sp.issparse(features):
            features = features.todense().A # make it easier for njit processing

        _A = torch.sparse.FloatTensor(edge_index, edge_weight, (features.shape[0],features.shape[0]))
        adj = torch_sparse_to_scipy_sparse(_A)
        adj_triu = sp.triu(adj, format='csr')

        if self.binary_feature:
            removed_cnt = dropedge_jaccard(adj_triu.data, adj_triu.indptr, adj_triu.indices, features, threshold=self.threshold)
        else:
            removed_cnt = dropedge_cosine(adj_triu.data, adj_triu.indptr, adj_triu.indices, features, threshold=self.threshold)
        
        # print('removed %s edges in the original graph' % removed_cnt)
        modified_adj = adj_triu + adj_triu.transpose()
        
        row, col = edge_index[0].cpu().numpy(), edge_index[1].cpu().numpy()
        adj_values = torch.tensor(modified_adj.toarray()[row,col], dtype=torch.float32).to(edge_index.device)

        return adj_values

        # elif isinstance(_A, Tensor):
        #     self.modified_adj = self.drop_dissimilar_edges(features, _A)
        #     self.check_dense_matrix()

        # return self.modified_adj

    def drop_dissimilar_edges(self, features, adj):
        """Drop dissimilar edges. (Slower version)
        """
        # preprocessing based on features
        edges = np.array(adj.nonzero().detach().cpu()).T
        removed_cnt = 0
        for edge in edges:
            n1 = edge[0]
            n2 = edge[1]
            if n1 > n2:
                continue

            if self.binary_feature:
                J = self._jaccard_similarity(features[n1], features[n2])

                if J < self.threshold:
                    adj[n1, n2] = 0
                    adj[n2, n1] = 0
                    removed_cnt += 1
            else:
                # For not binary feature, use cosine similarity
                C = self._cosine_similarity(features[n1], features[n2])
                if C < self.threshold:
                    adj[n1, n2] = 0
                    adj[n2, n1] = 0
                    removed_cnt += 1

        print('removed %s edges in the original graph' % removed_cnt)
        return adj

    def _jaccard_similarity(self, a, b):
        intersection = np.count_nonzero(np.multiply(a,b))
        if (np.count_nonzero(a) + np.count_nonzero(b) - intersection)==0:
            print(f'!Intersection=0!   a:{np.count_nonzero(a)}  b:{np.count_nonzero(b)}  intersection:{intersection}')
            with open('jaccard.txt','a') as f:
                f.write(f'{np.count_nonzero(a)}, {np.count_nonzero(b)}, {intersection}  \n')

            return 0
        J = intersection * 1.0 / (np.count_nonzero(a) + np.count_nonzero(b) - intersection)
        return J

    def _cosine_similarity(self, a, b):
        inner_product = (a * b).sum()
        C = inner_product / (np.sqrt(np.square(a).sum()) * np.sqrt(np.square(b).sum()) + 1e-10)
        return C


@njit
def dropedge_jaccard(A, iA, jA, features, threshold):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            a, b = features[n1], features[n2]
            intersection = np.count_nonzero(a*b)

            J = intersection * 1.0 / (np.count_nonzero(a) + np.count_nonzero(b) - intersection)

            if J < threshold:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1
    return removed_cnt


@njit
def dropedge_cosine(A, iA, jA, features, threshold):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            a, b = features[n1], features[n2]
            inner_product = (a * b).sum()
            C = inner_product / (np.sqrt(np.square(a).sum()) * np.sqrt(np.square(b).sum()) + 1e-8)

            if C < threshold:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1
    return removed_cnt




class GNNGuard(RobustIdentity):
    def __init__(self, use_gate=True, drop=False):
        super(GNNGuard, self).__init__()
        self.drop = drop
        self.use_gate = use_gate
        if self.use_gate:
            self.gate = Parameter(torch.rand(1)) # creat a generator between [0,1]


    def forward(self, edge_index, edge_weight, features):
        adj_values = self.att_coef(features, edge_index, i=0)
        # adj_2 = self.att_coef(x, adj, i=1)
        if self.use_gate:
            adj_values = self.gate * edge_weight + (1 - self.gate) * adj_values

        adj_values = torch.clamp(adj_values, min=0)
        
        return adj_values

    def att_coef(self, fea, edge_index, is_lil=False, i=0):
        # if is_lil == False:
        #     edge_index = edge_index._indices()
        # else:
        #     edge_index = edge_index.tocoo()

        n_node = fea.shape[0]
        row, col = edge_index[0].cpu().numpy()[:], edge_index[1].cpu().numpy()[:]

        fea_copy = fea.cpu().data.numpy()
        sim_matrix = cosine_similarity(X=fea_copy, Y=fea_copy)  # try cosine similarity
        # sim_matrix = torch.from_numpy(sim_matrix)
        sim = sim_matrix[row, col]
        sim[sim<0.1] = 0
        # print('dropped {} edges'.format(1-sim.nonzero()[0].shape[0]/len(sim)))

        # """use jaccard for binary features and cosine for numeric features"""
        # fea_start, fea_end = fea[edge_index[0]], fea[edge_index[1]]
        # isbinray = np.array_equal(fea_copy, fea_copy.astype(bool))  # check is the fea are binary
        # np.seterr(divide='ignore', invalid='ignore')
        # if isbinray:
        #     fea_start, fea_end = fea_start.T, fea_end.T
        #     sim = jaccard_score(fea_start, fea_end, average=None)  # similarity scores of each edge
        # else:
        #     fea_copy[np.isinf(fea_copy)] = 0
        #     fea_copy[np.isnan(fea_copy)] = 0
        #     sim_matrix = cosine_similarity(X=fea_copy, Y=fea_copy)  # try cosine similarity
        #     sim = sim_matrix[edge_index[0], edge_index[1]]
        #     sim[sim < 0.01] = 0

        """build a attention matrix"""
        att_dense = lil_matrix((n_node, n_node), dtype=np.float32)
        att_dense[row, col] = sim
        if att_dense[0, 0] == 1:
            att_dense = att_dense - sp.diags(att_dense.diagonal(), offsets=0, format="lil")
        # normalization, make the sum of each row is 1
        att_dense_norm = normalize(att_dense, axis=1, norm='l1')


        """add learnable dropout, make character vector"""
        if self.drop:
            character = np.vstack((att_dense_norm[row, col].A1,
                                     att_dense_norm[col, row].A1))
            character = torch.from_numpy(character.T)
            drop_score = self.drop_learn_1(character)
            drop_score = torch.sigmoid(drop_score)  # do not use softmax since we only have one element
            mm = torch.nn.Threshold(0.5, 0)
            drop_score = mm(drop_score)
            mm_2 = torch.nn.Threshold(-0.49, 1)
            drop_score = mm_2(-drop_score)
            drop_decision = drop_score.clone().requires_grad_()
            # print('rate of left edges', drop_decision.sum().data/drop_decision.shape[0])
            drop_matrix = lil_matrix((n_node, n_node), dtype=np.float32)
            drop_matrix[row, col] = drop_decision.cpu().data.numpy().squeeze(-1)
            att_dense_norm = att_dense_norm.multiply(drop_matrix.tocsr())  # update, remove the 0 edges

        if att_dense_norm[0, 0] == 0:  # add the weights of self-loop only add self-loop at the first layer
            degree = (att_dense_norm != 0).sum(1).A1
            # degree = degree.squeeze(-1).squeeze(-1)
            lam = 1 / (degree + 1) # degree +1 is to add itself
            self_weight = sp.diags(np.array(lam), offsets=0, format="lil")
            att = att_dense_norm + self_weight  # add the self loop
        else:
            att = att_dense_norm

        att_adj = edge_index
        att_edge_weight = att[row, col]
        att_edge_weight = np.exp(att_edge_weight)   # exponent, kind of softmax
        att_edge_weight = torch.tensor(np.array(att_edge_weight)[0], dtype=torch.float32).cuda()

        # shape = (n_node, n_node)
        # new_adj = torch.sparse.FloatTensor(att_adj, att_edge_weight, shape)
        return att_edge_weight

    def add_loop_sparse(self, adj, fill_value=1):
        # make identify sparse tensor
        row = torch.range(0, int(adj.shape[0]-1), dtype=torch.int64)
        i = torch.stack((row, row), dim=0)
        v = torch.ones(adj.shape[0], dtype=torch.float32)
        shape = adj.shape
        I_n = torch.sparse.FloatTensor(i, v, shape)
        return adj + I_n.to(self.device)


class VPN(RobustIdentity):
    def __init__(self, r=2):
        super(VPN, self).__init__()
        self.r=r 

        self.theta_1 = Parameter(torch.FloatTensor(1), requires_grad=True)
        self.theta_2 = Parameter(torch.FloatTensor(1), requires_grad=True)
        self.theta_1.data.fill_(1)
        self.theta_2.data.fill_(0)

        if r>=3:
            self.theta_3 = Parameter(torch.FloatTensor(1), requires_grad=True)
            self.theta_3.data.fill_(0)

        elif r==4:
            self.theta_4 = Parameter(torch.FloatTensor(1), requires_grad=True)
            self.theta_4.data.fill_(0)

        self.edge_index=None
        
    def preprocess_adj_alls(self, edge_index, edge_weight, features):
        # print('======VPN======')
        num_nodes = features.size(0)
        self.device=edge_index.device
        self.use_sparse_adj = True

        i = edge_index.cpu().numpy()
        sp_A = sp.csr_matrix((edge_weight.detach().cpu().numpy(), (i[0],i[1])), shape=(num_nodes,num_nodes))
        sp_A = sp_A + sp.eye(num_nodes)
        sp_A[sp_A > 1] = 1
        # A^(r)
        adj_alls =[sp_A]
        for k in range(2, self.r+1):
            adj_k = sp_A ** k
            adj_alls.append(adj_k)

        return adj_alls

    def forward(self, edge_index, edge_weight, features):
        # if (self.edge_index is None) or (edge_index.size()!=self.edge_index.size()) or (edge_index!=self.edge_index).sum()>0:
        #     self.edge_index = edge_index
        self.adj_alls = self.preprocess_adj_alls(edge_index, edge_weight, features)

        # sparsify
        adj_values = self.sparsification(edge_index, edge_weight, self.adj_alls, features.detach().cpu().numpy())
        
        
        #====== dense version
        # # add self loops
        # _A = _A - torch.diag_embed(torch.diag(_A)).to(self.device)
        # _A = _A + torch.diag_embed(torch.ones(num_nodes)).to(self.device)
        # sp_A = _A.to_sparse()

        # # A^(r)
        # adj_alls =[_A]
        # for i in range(1, self.r):
        #     adj_k = torch.sparse.mm(sp_A, adj_alls[-1])
        #     adj_k[adj_k>1]=1.
        #     adj_alls.append(adj_k)
        
        # # sparsify
        # self.modified_adj = self.sparsification(adj_alls, features.detach().cpu().numpy())
        # if not _A.is_sparse:
        #     self.check_dense_matrix()
        # print('VPN:',self.theta_1, self.theta_2)
        adj_values = torch.clamp(adj_values, min=0)
        return adj_values

    def sparsification(self, edge_index, edge_weight, adjs, X, sparse_rate=2.0, metric='euclidean'):
        """
        Parameters
        --------
        adjs: list of torch.Tensor dense matrix /scipy sparse
            [A^(k)]
        x: numpy.array

        Returns:
        --------
        modified_adj: torch 
            modified dense adjacency matrix
        """
        # if sp.issparse(adjs[0]):
        # scipy.sparse
        _A,_A_robust = adjs[0], adjs[-1]
        sparse_ids = generate_sparse_ids(_A, _A_robust, X, sparse_rate=sparse_rate, metric=metric)

        row, col = edge_index[0].cpu().numpy(), edge_index[1].cpu().numpy()
        n_nodes = adjs[0].shape[0]

        adj_values = self.theta_1*edge_weight 
        if self.r>=2:
            adj_k = (adjs[1]-adjs[0]).tolil()
            adj_k.setdiag(np.zeros(n_nodes))
            for u in range(n_nodes):
                adj_k[u,sparse_ids[u]]=0
            adj_values = adj_values + self.theta_2 * torch.tensor(adj_k.toarray()[row,col], dtype=torch.float32).to(self.device)
        if self.r>=3:
            adj_k = (adjs[2]-adjs[1]).tolil()
            adj_k.setdiag(np.zeros(n_nodes))
            for u in range(n_nodes):
                adj_k[u,sparse_ids[u]]=0
            adj_values = adj_values + self.theta_3 * torch.tensor(adj_k.toarray()[row,col], dtype=torch.float32).to(self.device)

        # else:
        #     modified_adj = torch.from_numpy(adjs[0].todense()).to(self.device)* self.theta_1
        #     if self.r>=2:
        #         modified_adj += torch.from_numpy(adjs[1].todense()).to(self.device) * self.theta_2
        #     elif self.r>=3:
        #         modified_adj += torch.from_numpy(adjs[2].todense()).to(self.device) * self.theta_3
        #     elif self.r>=4:
        #         modified_adj += torch.from_numpy(adjs[3].todense()).to(self.device) * self.theta_4

        # print('modified adj VPN:', modified_adj)

        return adj_values

def generate_sparse_ids(_A, _A_robust, X, sparse_rate, metric):
    '''
    _A: origin adjacency matrix , sp.csr_matrix
    X: feature matrix. numpy.array
    return: list n
    '''
    sparse_ids = []
    _A.setdiag(np.zeros(_A.shape[0]))
    _D = np.count_nonzero(_A.toarray(), axis=1)

    _A_robust.setdiag(np.zeros(_A_robust.shape[0]))
    _D_r = np.count_nonzero(_A_robust.toarray(), axis=1)

    d_mean = np.mean(_D)
    d_std = np.std(_D)
    if metric=='correlation':
        d_thres = d_mean+2.5*d_std 
    else:
        d_thres = d_mean+2*d_std 
    highD,sparseN,nosparseN = 0,0,0

    for u in range(X.shape[0]):
        neighbors = _A_robust[u,:].nonzero()[1]
        x_neighbors = X[neighbors]
        x_u = X[u]
        
        if _D[u] > d_thres: # high degree nodes, do not power at all
            sparse_t = np.setdiff1d(np.arange(len(_D)), _A[u,:].nonzero()[1])
            sparse_ids.append(sparse_t)
            highD += 1
        elif round(sparse_rate*_D[u]) > _D_r[u]: # sparse rate too high, power all
            sparse_ids.append(np.array([]))
            nosparseN += 1
        else:
            if metric=='correlation':
                dist_2 = np.squeeze(scipy.spatial.distance.cdist(x_u,x_neighbors,'correlation'))
            else:
                dist_2 = np.sum((x_neighbors - x_u) ** 2, axis=1)
        
            nz_sel = int(_D_r[u]-round(sparse_rate*_D[u]))
            sparse_ids.append(neighbors[dist_2.argsort()[-nz_sel:]])
            sparseN += 1
        
    # print('highD {0}, sparified {1}, no sparsified {2}'.format(highD,sparseN,nosparseN))
    return sparse_ids
        

class LambdaModule(nn.Module):
    def __init__(self, lambd):
        super().__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.lambd)
def act_map_nn(act):
    return LambdaModule(act_map(act))


GRAPHNAS_DEFAULT_ACT_OPS = [
    # "sigmoid", "tanh", "relu", "linear",
    #  "softplus", "leaky_relu", "relu6", "elu"
    "sigmoid",
    "tanh",
    "relu",
    "linear",
    "elu",
]

@register_nas_space("grnaspace")
class GRNASpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.6,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = ['gcn', "gat_2"],
        # rob_ops: _typ.Tuple = ["identity","svd","jaccard","gnnguard", "vpn"]
        rob_ops: _typ.Tuple = ["identity","svd","jaccard","gnnguard"],
        act_ops: _typ.Sequence[_typ.Union[str, _typ.Any]] = GRAPHNAS_DEFAULT_ACT_OPS,
    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.ops = ops
        self.rob_ops = rob_ops
        self.dropout = dropout
        self.act_ops = act_ops

    def instantiate(
        self,
        hidden_dim: _typ.Optional[int] = None,
        layer_number: _typ.Optional[int] = None,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = None,
        rob_ops: _typ.Tuple = None,
        act_ops: _typ.Tuple = None,
        dropout=None,
    ):
        super().instantiate()
        self.hidden_dim = hidden_dim or self.hidden_dim
        self.layer_number = layer_number or self.layer_number
        self.input_dim = input_dim or self.input_dim
        self.output_dim = output_dim or self.output_dim
        self.ops = ops or self.ops
        self.rob_ops = rob_ops or self.rob_ops
        self.act_ops = act_ops or self.act_ops
        self.dropout = dropout or self.dropout
        for layer in range(self.layer_number):
            setattr(
                self,
                f"op_{layer}",
                self.setLayerChoice(
                    layer,
                    [
                        op(
                            self.input_dim if layer == 0 else self.hidden_dim,
                            self.output_dim
                            if layer == self.layer_number - 1
                            else self.hidden_dim,
                        )
                        if isinstance(op, type)
                        else gnn_map(
                            op,
                            self.input_dim if layer == 0 else self.hidden_dim,
                            self.output_dim
                            if layer == self.layer_number - 1
                            else self.hidden_dim,
                        )
                        for op in self.ops
                    ],
                ),
            )

            setattr(
                self,
                f"rob_op_{layer}",

                self.setLayerChoice(
                    layer,
                    [
                        op()
                        if isinstance(op, type)
                        else ROB_OPS[op]()
                        for op in self.rob_ops
                    ],
                ),
                
            )
            setattr(
            self,
            "act",
            self.setLayerChoice(
                2 * layer, [act_map_nn(a) for a in self.act_ops], key="act"
            ),
        )
        self._initialized = True

    def forward(self, data):
        x = bk_feat(data)
        edge_weight = data.edge_weight if data.edge_weight is not None else torch.ones(data.edge_index.size(1)).to(data.edge_index.device)
        for layer in range(self.layer_number):
            rob_op = getattr(self, f"rob_op_{layer}")
            edge_weight = rob_op(data.edge_index, edge_weight, x)
            op = getattr(self, f"op_{layer}")
            x = op(x, data.edge_index, edge_weight)
            if layer != self.layer_number - 1:
                # x = F.leaky_relu(x)
                act = getattr(self, "act")
                x = act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return F.log_softmax(x, dim=1)

    def parse_model(self, selection, device) -> BaseAutoModel:
        # return AutoGCN(self.input_dim, self.output_dim, device)
        return self.wrap().fix(selection)


import os
import os.path as osp
import shutil
import pickle
import numpy as np
import torch
import typing as _typing

from autogl.data import Data, download_url, InMemoryStaticGraphSet
from autogl.data.graph import GeneralStaticGraphGenerator
from ._dataset_registry import DatasetUniversalRegistry
from ._data_source import OnlineDataSource
from .. import backend as _backend


def _untar(path, fname, delete_tar=True):
    """
    Unpacks the given archive file to the same directory, then (by default)
    deletes the archive file.
    """
    print("unpacking " + fname)
    full_path = os.path.join(path, fname)
    shutil.unpack_archive(full_path, path)
    if delete_tar:
        os.remove(full_path)


class _GTNDataSource(OnlineDataSource):
    def __init__(self, path: str, name: str):
        self.__name: str = name
        self.__url: str = (
            f"https://github.com/cenyk1230/gtn-data/blob/master/{name}.zip?raw=true"
        )
        super(_GTNDataSource, self).__init__(path)
        self.__data = torch.load(list(self._processed_file_paths)[0])

    @property
    def _raw_filenames(self) -> _typing.Iterable[str]:
        return ["edges.pkl", "labels.pkl", "node_features.pkl"]

    @property
    def _processed_filenames(self) -> _typing.Iterable[str]:
        return ["data.pt"]

    def __read_gtn_data(self, directory):
        edges = pickle.load(open(osp.join(directory, "edges.pkl"), "rb"))
        labels = pickle.load(open(osp.join(directory, "labels.pkl"), "rb"))
        node_features = pickle.load(open(osp.join(directory, "node_features.pkl"), "rb"))

        data = Data()
        data.x = torch.from_numpy(node_features).float()

        num_nodes = edges[0].shape[0]

        node_type = np.zeros(num_nodes, dtype=int)
        assert len(edges) == 4
        assert len(edges[0].nonzero()) == 2

        node_type[edges[0].nonzero()[0]] = 0
        node_type[edges[0].nonzero()[1]] = 1
        node_type[edges[1].nonzero()[0]] = 1
        node_type[edges[1].nonzero()[1]] = 0
        node_type[edges[2].nonzero()[0]] = 0
        node_type[edges[2].nonzero()[1]] = 2
        node_type[edges[3].nonzero()[0]] = 2
        node_type[edges[3].nonzero()[1]] = 0

        print(node_type)
        data.pos = torch.from_numpy(node_type)

        edge_list = []
        for i, edge in enumerate(edges):
            edge_tmp = torch.from_numpy(
                np.vstack((edge.nonzero()[0], edge.nonzero()[1]))
            ).long()
            edge_list.append(edge_tmp)
        data.edge_index = torch.cat(edge_list, 1)

        A = []
        for i, edge in enumerate(edges):
            edge_tmp = torch.from_numpy(
                np.vstack((edge.nonzero()[0], edge.nonzero()[1]))
            ).long()
            value_tmp = torch.ones(edge_tmp.shape[1]).float()
            A.append((edge_tmp, value_tmp))
        edge_tmp = torch.stack(
            (torch.arange(0, num_nodes), torch.arange(0, num_nodes))
        ).long()
        value_tmp = torch.ones(num_nodes).float()
        A.append((edge_tmp, value_tmp))
        data.adj = A

        data.train_node = torch.from_numpy(np.array(labels[0])[:, 0]).long()
        data.train_target = torch.from_numpy(np.array(labels[0])[:, 1]).long()
        data.valid_node = torch.from_numpy(np.array(labels[1])[:, 0]).long()
        data.valid_target = torch.from_numpy(np.array(labels[1])[:, 1]).long()
        data.test_node = torch.from_numpy(np.array(labels[2])[:, 0]).long()
        data.test_target = torch.from_numpy(np.array(labels[2])[:, 1]).long()

        y = np.zeros(num_nodes, dtype=int)
        x_index = torch.cat((data.train_node, data.valid_node, data.test_node))
        y_index = torch.cat((data.train_target, data.valid_target, data.test_target))
        y[x_index.numpy()] = y_index.numpy()
        data.y = torch.from_numpy(y)
        self.__data = data

    def __transform_gtn_data(self):
        self.__data.train_mask = torch.zeros(self.__data.x.size(0), dtype=torch.bool)
        self.__data.val_mask = torch.zeros(self.__data.x.size(0), dtype=torch.bool)
        self.__data.test_mask = torch.zeros(self.__data.x.size(0), dtype=torch.bool)
        self.__data.train_mask[getattr(self.__data, "train_node")] = True
        self.__data.val_mask[getattr(self.__data, "valid_node")] = True
        self.__data.test_mask[getattr(self.__data, "test_node")] = True

    def _fetch(self):
        download_url(self.__url, self._raw_directory, name=f"{self.__name}.zip")
        _untar(self._raw_directory, f"{self.__name}.zip")

    def _process(self):
        self.__read_gtn_data(self._raw_directory)
        self.__transform_gtn_data()
        torch.save(self.__data, list(self._processed_file_paths)[0])

    def __len__(self) -> int:
        return 1

    def __getitem__(self, index):
        if index != 0:
            raise IndexError
        return self.__data


@DatasetUniversalRegistry.register_dataset("gtn-acm")
class GTNACMDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        data = _GTNDataSource(path, "gtn-acm")[0]
        if _backend.DependentBackend.is_dgl():
            super(GTNACMDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'feat': getattr(data, 'x'),
                            'label': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )
        elif _backend.DependentBackend.is_pyg():
            super(GTNACMDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'x': getattr(data, 'x'),
                            'y': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )


@DatasetUniversalRegistry.register_dataset("gtn-dblp")
class GTNDBLPDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        data = _GTNDataSource(path, "gtn-dblp")[0]
        if _backend.DependentBackend.is_dgl():
            super(GTNDBLPDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'feat': getattr(data, 'x'),
                            'label': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )
        elif _backend.DependentBackend.is_pyg():
            super(GTNDBLPDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'x': getattr(data, 'x'),
                            'y': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )


@DatasetUniversalRegistry.register_dataset("gtn-imdb")
class GTNIMDBDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        data = _GTNDataSource(path, "gtn-imdb")[0]
        if _backend.DependentBackend.is_dgl():
            super(GTNIMDBDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'feat': getattr(data, 'x'),
                            'label': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )
        elif _backend.DependentBackend.is_pyg():
            super(GTNIMDBDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {
                            'x': getattr(data, 'x'),
                            'y': getattr(data, 'y'),
                            'pos': getattr(data, 'pos'),
                            'train_mask': getattr(data, 'train_mask'),
                            'val_mask': getattr(data, 'val_mask'),
                            'test_mask': getattr(data, 'test_mask')
                        },
                        getattr(data, 'edge_index')
                    )
                ]
            )
